# flake8: noqa
import numpy as np

class DiffusionCoefficient:

    def __init__(self, traj, timestep, atom_indices=None, molecule=False):
        """

        This class calculates the Diffusion Coefficient for the given Trajectory using the Einstein Equation:
        
        :math: `\\left \\langle  \\left | r(t) - r(0) \\right | ^{2} \\right \\rangle  = 2nDt`
        
        where r(t) is the position of atom at time t, n is the degrees of freedom and D is the Diffusion Coefficient
        
        Solved herein by fitting with :math: `y = mx + c`, i.e. :math: `\\frac{1}{2n} \\left \\langle  \\left | r(t) - r(0) \\right | ^{2} \\right \\rangle  = Dt`, with m = D and c = 0

        wiki : https://en.wikibooks.org/wiki/Molecular_Simulation/Diffusion_Coefficients

        Parameters:
            traj (Trajectory): 
                Trajectory of atoms objects (images)
            timestep (Float): 
                Timestep between *each image in the trajectory*, in ASE timestep units
                (For an MD simulation with timestep of N, and images written every M iterations, our timestep here is N * M  
            atom_indices (List of Int): 
                The indices of atoms whose Diffusion Coefficient is to be calculated explicitly
            molecule (Boolean)
                Indicate if we are studying a molecule instead of atoms, therefore use centre of mass in calculations
 
        """

        self.traj = traj
        self.timestep = timestep

        # Condition used if user wants to calculate diffusion coefficients for specific atoms or all atoms
        self.atom_indices = atom_indices
        if self.atom_indices == None:
            self.atom_indices = [i for i in range(len(traj[0]))] 

        # Condition if we are working with the mobility of a molecule, need to manage arrays slightly differently
        self.is_molecule = molecule
        if self.is_molecule:
            self.types_of_atoms = ["molecule"]
            self.no_of_atoms = [1]
        else:
            self.types_of_atoms = sorted(set(traj[0].symbols[self.atom_indices]))
            self.no_of_atoms = [traj[0].get_chemical_symbols().count(symbol) for symbol in self.types_of_atoms]

        # Dummy initialisation for important results data object
        self._slopes = []

    @property
    def no_of_types_of_atoms(self):
        """

        Dynamically returns the number of different atoms in the system
        
        """
        return len(self.types_of_atoms)

    @property
    def slopes(self):
        """

        Method to return slopes fitted to datapoints. If undefined, calculate slopes

        """
        if len(self._slopes) == 0:
            self.calculate()
        return self._slopes

    @slopes.setter
    def slopes(self, values):
        """
 
        Method to set slopes as fitted to datapoints

        """
        self._slopes = values

    def _initialise_arrays(self, ignore_n_images, number_of_segments):
        """

        Private function to initialise data storage objects. This includes objects to store the total timesteps
        sampled, the average diffusivity for species in any given segment, and objects to store gradient and intercept from fitting.

        Parameters:
            ignore_n_images (Int): 
                Number of images you want to ignore from the start of the trajectory, e.g. during equilibration
            number_of_segments (Int): 
                Divides the given trajectory in to segments to allow statistical analysis
        
        """

        total_images = len(self.traj) - ignore_n_images
        self.no_of_segments = number_of_segments
        self.len_segments = total_images // self.no_of_segments

        # These are the data objects we need when plotting information. First the x-axis, timesteps
        self.timesteps = np.linspace(0,total_images*self.timestep,total_images+1)
        # This holds all the data points for the diffusion coefficients, averaged over atoms
        self.xyz_segment_ensemble_average = np.zeros((self.no_of_segments,self.no_of_types_of_atoms,3,self.len_segments))
        # This holds all the information on linear fits, from which we get the diffusion coefficients
        self.slopes = np.zeros((self.no_of_types_of_atoms,self.no_of_segments,3))
        self.intercepts = np.zeros((self.no_of_types_of_atoms,self.no_of_segments,3))

        self.cont_xyz_segment_ensemble_average = 0

    def calculate(self, ignore_n_images=0, number_of_segments=1):
        """
        
        Calculate the diffusion coefficients, using the previously supplied data. The user can break the data into segments and 
        take the average over these trajectories, therefore allowing statistical analysis and derivation of standard deviations.
        Option is also provided to ignore initial images if data is perhaps unequilibrated initially.

        Parameters:
            ignore_n_images (Int): 
                Number of images you want to ignore from the start of the trajectory, e.g. during equilibration
            number_of_segments (Int): 
                Divides the given trajectory in to segments to allow statistical analysis

        """

        # Setup all the arrays we need to store information
        self._initialise_arrays(ignore_n_images, number_of_segments)

        for segment_no in range(self.no_of_segments):
            start = segment_no*self.len_segments  
            end = start + self.len_segments
            seg = self.traj[ignore_n_images+start:ignore_n_images+end]

            # If we are considering a molecular system, work out the COM for the starting structure
            if self.is_molecule:
                com_orig = np.zeros(3)
                for atom_no in self.atom_indices:
                    com_orig += seg[0].positions[atom_no] / len(self.atom_indices)

            # For each image, calculate displacement.
            # I spent some time deciding if this should run from 0 or 1, as the displacement will be zero for 
            # t = 0, but this is a data point that needs fitting too and so should be included
            for image_no in range(0,len(seg)): 
                # This object collects the xyz displacements for all atom species in the image
                xyz_disp = np.zeros((self.no_of_types_of_atoms,3))
                
                # Calculating for each atom individually, grouping by species type (e.g. solid state)
                if not self.is_molecule:
                    # For each atom, work out displacement from start coordinate and collect information with like atoms
                    for atom_no in self.atom_indices:
                        sym_index = self.types_of_atoms.index(seg[image_no].symbols[atom_no])
                        xyz_disp[sym_index] += np.square(seg[image_no].positions[atom_no] - seg[0].positions[atom_no])
        
                else: # Calculating for group of atoms (molecule) and work out squared displacement
                    com_disp = np.zeros(3)
                    for atom_no in self.atom_indices:
                        com_disp += seg[image_no].positions[atom_no] / len(self.atom_indices)
                    xyz_disp[0] += np.square(com_disp - com_orig)

                # For each atom species or molecule, use xyz_disp to calculate the average data                      
                for sym_index in range(self.no_of_types_of_atoms):
                    # Normalise by degrees of freedom and average overall atoms for each axes over entire segment                         
                    denominator = (2*self.no_of_atoms[sym_index])
                    for xyz in range(3):
                        self.xyz_segment_ensemble_average[segment_no][sym_index][xyz][image_no] = (xyz_disp[sym_index][xyz]/denominator)

            # We've collected all the data for this entire segment, so now to fit the data.
            for sym_index in range(self.no_of_types_of_atoms):    
                self.slopes[sym_index][segment_no], self.intercepts[sym_index][segment_no] = self._fit_data(self.timesteps[start:end], 
                                                                                                            self.xyz_segment_ensemble_average[segment_no][sym_index])

    def _fit_data(self, x, y):
        """
        Private function that returns slope and intercept for linear fit to mean square diffusion data


        Parameters:
            x (Array of floats): 
                Linear list of timesteps in the calculation
            y (Array of floats): 
                Mean square displacement as a function of time.
        
        """

        # Simpler implementation but disabled as fails Conda tests.
        # from scipy.stats import linregress
        # slope, intercept, r_value, p_value, std_err = linregress(x,y)
       
        # Initialise objects
        slopes = np.zeros(3)
        intercepts = np.zeros(3)

        # Convert into suitable format for lstsq
        x_edited = np.vstack([np.array(x), np.ones(len(x))]).T
        # Calculate slopes for x, y and z-axes
        for xyz in range(3):
            slopes[xyz], intercepts[xyz] = np.linalg.lstsq(x_edited, y[xyz], rcond=-1)[0]

        return slopes, intercepts

    def get_diffusion_coefficients(self):
        """
        
        Returns diffusion coefficients for atoms (in alphabetical order) along with standard deviation.
        
        All data is currently passed out in units of Å^2/<ASE time units>
        To convert into Å^2/fs => multiply by ase.units.fs
        To convert from Å^2/fs to cm^2/s => multiply by (10^-8)^2 / 10^-15 = 10^-1
        
        """
       
        slopes = [np.mean(self.slopes[sym_index]) for sym_index in range(self.no_of_types_of_atoms)]
        std = [np.std(self.slopes[sym_index]) for sym_index in range(self.no_of_types_of_atoms)]
 
        return slopes, std

    def plot(self, ax=None, show=False):
        """
        
        Auto-plot of Diffusion Coefficient data. Provides basic framework for visualising analysis.
 
         Parameters:
            ax (Matplotlib.axes.Axes)
                Axes object on to which plot can be created
            show (Boolean)
                Whether or not to show the created plot. Default: False
        
        """

        # Necessary if user hasn't supplied an axis. 
        import matplotlib.pyplot as plt

        # Convert from ASE time units to fs (aesthetic)
        from ase.units import fs as fs_conversion
        
        if ax is None:
            ax = plt.gca()

        # Define some aesthetic variables
        color_list = plt.cm.Set3(np.linspace(0, 1, self.no_of_types_of_atoms))
        xyz_labels=['X','Y','Z']
        xyz_markers = ['o','s','^']

        # Create an x-axis that is in a more intuitive format for the view
        graph_timesteps = self.timesteps / fs_conversion
        
        for segment_no in range(self.no_of_segments):
            start = segment_no*self.len_segments  
            end = start + self.len_segments
            label = None
            
            for sym_index in range(self.no_of_types_of_atoms): 
                for xyz in range(3):
                    if segment_no == 0:
                        label = 'Species: %s (%s)'%(self.types_of_atoms[sym_index], xyz_labels[xyz])
                    # Add scatter graph  for the mean square displacement data in this segment
                    ax.scatter(graph_timesteps[start:end], self.xyz_segment_ensemble_average[segment_no][sym_index][xyz],
                             color=color_list[sym_index], marker=xyz_markers[xyz], label=label, linewidth=1, edgecolor='grey')

                # Print the line of best fit for segment      
                line = np.mean(self.slopes[sym_index][segment_no])*fs_conversion*graph_timesteps[start:end]+np.mean(self.intercepts[sym_index][segment_no])
                if segment_no == 0:
                    label = 'Segment Mean : %s'%(self.types_of_atoms[sym_index])
                ax.plot(graph_timesteps[start:end], line, color='C%d'%(sym_index), label=label, linestyle='--')
 
            # Plot separator at end of segment
            x_coord = graph_timesteps[end-1]
            ax.plot([x_coord, x_coord],[-0.001, 1.05*np.amax(self.xyz_segment_ensemble_average)], color='grey', linestyle=":")

        # Plot the overall mean (average of slopes) for each atom species
        # This only makes sense if the data is all plotted on the same x-axis timeframe, which currently we are not - everything is plotted sequentially
        #for sym_index in range(self.no_of_types_of_atoms):
        #    line = np.mean(self.slopes[sym_index])*graph_timesteps+np.mean(self.intercepts[sym_index])
        #    label ='Mean, Total : %s'%(self.types_of_atoms[sym_index])
        #    ax.plot(graph_timesteps, line, color='C%d'%(sym_index), label=label, linestyle="-")

        # Aesthetic parts of the plot
        ax.set_ylim(-0.001, 1.05*np.amax(self.xyz_segment_ensemble_average))
        ax.legend(loc='best')
        ax.set_xlabel('Time (fs)')
        ax.set_ylabel(r'Mean Square Displacement ($\AA^2$)')

        if show:
            plt.show()

    def print_data(self):
        """
        
        Output of statistical analysis for Diffusion Coefficient data. Provides basic framework for understanding calculation.
        
        """

        from ase.units import fs as fs_conversion

        # Collect statistical data for diffusion coefficient over all segments
        slopes, std = self.get_diffusion_coefficients()

        # Useful notes for any consideration of conversion. 
        # Converting gradient from Å^2/fs to more common units of cm^2/s => multiplying by (10^-8)^2 / 10^-15 = 10^-1
        # Converting intercept from Å^2 to more common units of cm^2 => multiply by (10^-8)^2 = 10^-16
        #
        # Note currently in ASE internal time units
        # Converting into fs => divide by 1/(fs_conversion) => multiply by (fs_conversion)

        # Print data for each atom, in each segment.
        for sym_index in range(self.no_of_types_of_atoms):
            print('---')
            print(r'Species: %4s' % self.types_of_atoms[sym_index])
            print('---')
            for segment_no in range(self.no_of_segments):
                print(r'Segment   %3d:         Diffusion Coefficient = %.10f Å^2/fs; Intercept = %.10f Å^2;' % 
                     (segment_no, np.mean(self.slopes[sym_index][segment_no])*fs_conversion, np.mean(self.intercepts[sym_index][segment_no])))

        # Print average overall data.
        print('---')
        for sym_index in range(self.no_of_types_of_atoms):
            print('Mean Diffusion Coefficient (X, Y and Z) : %s = %.10f Å^2/fs; Std. Dev. = %.10f Å^2/fs' % 
                 (self.types_of_atoms[sym_index], slopes[sym_index]*fs_conversion, std[sym_index]*fs_conversion))
        print('---')



# flake8: noqa
import numpy as np

import logging
log = logging.getLogger('Jacapo')

import ase.dft.kpoints
from ase.utils import basestring


'''
provides functions to determine if an input parameter has changed.
'''

#######################################################################
#### changed functions

def kpts_changed(calc, x):
    '''
    check if kpt grid has changed.

    we have to take care to generate the right k-points from x if
    needed. if a user provides (4,4,4) we need to generate the MP
    grid, etc...

    Since i changed the MP code in set_kpts, there is some
    incompatibility with old jacapo calculations and their MP
    grids.
    '''
    #chadi-cohen
    if isinstance(x, basestring):
        listofkpts = getattr(ase.dft.kpoints, x)
    #monkhorst-pack grid
    elif np.array(x).shape == (3,):
        from ase.dft.kpoints import monkhorst_pack
        N1, N2, N3 = x
        listofkpts = monkhorst_pack((N1, N2, N3))
    #user-defined list is provided
    elif len(np.array(x).shape) == 2:
        listofkpts = np.array(x)
    else:
        raise Exception('apparent invalid setting for kpts')

    grid = calc.get_kpts()

    if grid.shape != listofkpts.shape:
        return True

    if (abs(listofkpts - grid) < 1e-6).all():
        return False
    else:
        return True

def electronic_minimization_changed(calc, x):
    myx = calc.get_electronic_minimization()

    for key in myx:
        if myx[key] != x[key]:
            print(key, myx[key], ' changed to ', x[key])
            return True
    return False

def spinpol_changed(calc, x):
    if x != calc.get_spinpol():
        return True
    else:
        return False

def symmetry_changed(calc, x):
    if x != calc.get_symmetry():
        return True
    else:
        return False

def xc_changed(calc, x):
    if x != calc.get_xc():
        return True
    return False

def calculate_stress_changed(calc, x):
    if x != calc.get_calculate_stress():
        return True
    return False

def ados_changed(calc, x):
    ados = calc.get_ados()

    #ados may not be defined, and then None is returned
    if ados is None and x is None:
        return False
    elif ados is None and x is not None:
        return True
    elif ados is not None and x is None:
        return True

    #getting here means ados and x are not none so we compare them
    for key in x:
        try:
            if x[key] != ados[key]:
                return True
        except ValueError:
            if (x[key] != ados[key]).all():
                return True
    return False

def convergence_changed(calc, x):
    conv = calc.get_convergence()
    for key in x:
        if x[key] != conv[key]:
            return True
    return False

def charge_mixing_changed(calc, x):
    cm = calc.get_charge_mixing()
    if x is None and cm is None:
        return False
    else:
        return True

    for key in x:
        if x[key] != cm[key]:
            return True
    return False

def decoupling_changed(calc, x):
    pars = calc.get_decoupling()
    for key in x:
        if x[key] != pars[key]:
            return True
    return False

def dipole_changed(calc, x):

    pars = calc.get_dipole() #pars stored in calculator

    # pars = False if no dipole variables exist
    # XXX fix the "logic" below.   "if not pars and not x"
    # or can it be very different types?  This is very difficult to read
    if (pars is False and x is False):
        return False #no change
    elif (pars is False and x is not False):
        return True

    # both x and pars is a dictionary
    # XXX wtf?  type(dict) == dict.  If x is a dict then type(x) is at least
    if (isinstance(pars, type(dict)) and
        isinstance(pars, type(x))):
        for key in x:
            if key == 'position':    # dipole layer position is never written to the nc file
                print('need to do something special')
                continue
            if x[key] != pars[key]:
                return True

    #nothing seems to have changed.
    return False

def extpot_changed(calc, x):
    extpot = calc.get_extpot()
    if (x == extpot).all():
        return False
    return True

def fftgrid_changed(calc, x):
    # validkeys = ['soft', 'hard']

    myx = calc.get_fftgrid()
    if (myx['soft'] == x['soft'] and myx['hard'] == x['hard']):
        return False
    else:
        return True

def ncoutput_changed(calc, x):
    ncout = calc.get_ncoutput()
    for key in x:
        if x[key] != ncout[key]:
            return True
    return False

def nbands_changed(calc, x):
    if calc.get_nbands() == x:
        return False
    else:
        return True

def occupationstatistics_changed(calc, x):
    if calc.get_occupationstatistics() == x:
        return False
    else:
        return True

def pw_changed(calc, x):
    if calc.get_pw() == x:
        return False
    else:
        return True

def dw_changed(calc, x):
    if calc.get_dw() == x:
        return False
    else:
        return True

def ft_changed(calc, x):
    if calc.get_ft() == x:
        return False
    else:
        return True

def mdos_changed(calc,x):

    myx = calc.get_mdos()

    log.debug('myx = %s' % str(myx))
    log.debug('x = %s' % str(x))

    if x is None and myx is None:
        return False
    elif ((x is None and myx is not None)
        or (x is not None and myx is None)):
        return True
    else:
        for key in x:
            if x[key] != myx[key]:
                return True
    return False

def pseudopotentials_changed(calc,x):

    mypsp = calc.get_pseudopotentials()

    if len(mypsp) != len(x):
        return True

    for key in x:
        if key not in mypsp:
            return True
        if mypsp[key] != x[key]:
            return True

    for key in mypsp:
        if key not in x:
            return True
        if mypsp[key] != x[key]:
            return True
    return False

def status_changed(calc,x):
    if calc.get_status() != x:
        return True
    return False



# flake8: noqa
#!/usr/bin/env python3
'''
interface to symmol


unzip symmol.tar.tz
compile symmol if necessary:
g77 -o symmol symmol.f

make sure symmol is on your executable path
'''

import math,os,re,string

import numpy as np

from Scientific.Geometry import Vector
from Scientific.IO.FortranFormat import FortranLine, FortranFormat

class SYMMOL:
    def __init__(self,atoms,outfile=None):

        unitcell = atoms.get_cell()
        A = Vector(unitcell[0])
        B = Vector(unitcell[1])
        C = Vector(unitcell[2])

        # lengths of the vectors
        a = A.length()#*angstroms2bohr
        b = B.length()#*angstroms2bohr
        c = C.length()#*angstroms2bohr

        # angles between the vectors
        rad2deg = 360./(2.*math.pi)
        alpha = B.angle(C)*rad2deg
        beta = A.angle(C)*rad2deg
        gamma = A.angle(B)*rad2deg

        # scaledpositions = atoms.get_scaled_positions()
        # chemicalsymbols = [atom.get_symbol() for atom in atoms]

        input = ''
        input += '%1.3f %1.3f %1.3f %1.3f %1.3f %1.3f\n' % (a,b,c,
                                                            alpha,beta,gamma)
        input += '1 1 0.1 0.1\n'

        for atom in atoms:
            sym = atom.get_symbol()
            group = 1
            x,y,z = atom.get_position()
            #format(a6,i2,6f9.5)
            input += str(FortranLine((sym,
                                      group,
                                      x,y,z),
                                     FortranFormat('a6,i2,3f9.5')))+'\n'

        pin,pout = os.popen2('symmol')
        pin.writelines(input)
        pin.close()
        self.output = pout.readlines()
        pout.close()

        if outfile:
            f = open(outfile,'w')
            f.writelines(self.output)
            f.close()

        if os.path.exists('symmol.log'):
            os.remove('symmol.log')

    def __str__(self):
        return string.join(self.output)

    def get_point_group(self):
        regexp = re.compile('^ Schoenflies symbol')

        for line in self.output:
            if regexp.search(line):
                return line

    def get_moments_of_inertia(self):
        regexp = re.compile('^ PRINCIPAL INERTIA MOMENTS and DEGENERATION DEGREE')
        lines = open('symmol.out').readlines()
        for i,line in enumerate(lines):

            if regexp.search(line):
                data = lines[i+1]
                break
        [Ia, Ib, Ic, degen] =  data.split()
        return [float(Ia), float(Ib), float(Ic), int(degen)]

    def get_symmetry_operators(self):
        # regexp = re.compile(' SYMMETRY GROUP MATRICES')
        reg2 = re.compile(r'^  \d+ CSM')

        lines = open('symmol.out').readlines()

        matrices = []
        types = []

        for i,line in enumerate(lines):
            if reg2.search(line):
                fields = line.split()
                types.append(fields[-1]) #type is at the end

                row1 = [float(x) for x in lines[i+1].split()]
                row2 = [float(x) for x in lines[i+2].split()]
                row3 = [float(x) for x in lines[i+3].split()]

                matrices.append(np.array([row1, row2, row3]))

        return (matrices, types)



if __name__ == '__test__':

    from ase import Atoms
    from ase.data import molecules

    mol = 'NH3'
    atoms = Atoms(mol,
                positions = molecules.data[mol]['positions'])

    sg = SYMMOL(atoms)
    print(sg.get_point_group())
    print(sg.get_moments_of_inertia())
    print(atoms.get_moments_of_inertia())
    print(sg.get_symmetry_operators())

if __name__ == '__main__':
    from ase.calculators.jacapo import Jacapo
    from optparse import OptionParser

    parser = OptionParser(usage='symmol.py ncfile',
                      version='0.1')

    parser.add_option('-f',
                      nargs=0,
                      help = 'print full output')

    parser.add_option('-o',
                      nargs=1,
                      help = 'save output in filename')

    options,args = parser.parse_args()

    for ncfile in args:

        sy = SYMMOL(Jacapo.read_atoms(ncfile),outfile=options.o)

        print('Point group = ',sy.get_point_group())
        print('Moments of inertia = ',sy.get_moments_of_inertia())
        print('Symmetry operators = ',sy.get_symmetry_operators())
        if options.f is not None:
            print(sy)





"""Helper functions for creating the most common surfaces and related tasks.

The helper functions can create the most common low-index surfaces,
add vacuum layers and add adsorbates.

"""

from math import sqrt
from operator import itemgetter

import numpy as np

from ase.atom import Atom
from ase.atoms import Atoms
from ase.data import reference_states, atomic_numbers
from ase.lattice.cubic import FaceCenteredCubic
from ase.utils import basestring


def fcc100(symbol, size, a=None, vacuum=None, orthogonal=True,
           periodic=False):
    """FCC(100) surface.

    Supported special adsorption sites: 'ontop', 'bridge', 'hollow'."""
    if not orthogonal:
        raise NotImplementedError("Can't do non-orthogonal cell yet!")

    return _surface(symbol, 'fcc', '100', size, a, None, vacuum,
                    periodic=periodic,
                    orthogonal=orthogonal)


def fcc110(symbol, size, a=None, vacuum=None, orthogonal=True,
           periodic=False):
    """FCC(110) surface.

    Supported special adsorption sites: 'ontop', 'longbridge',
    'shortbridge', 'hollow'."""
    if not orthogonal:
        raise NotImplementedError("Can't do non-orthogonal cell yet!")

    return _surface(symbol, 'fcc', '110', size, a, None, vacuum,
                    periodic=periodic,
                    orthogonal=orthogonal)



def bcc100(symbol, size, a=None, vacuum=None, orthogonal=True,
           periodic=False):
    """BCC(100) surface.

    Supported special adsorption sites: 'ontop', 'bridge', 'hollow'."""
    if not orthogonal:
        raise NotImplementedError("Can't do non-orthogonal cell yet!")

    return _surface(symbol, 'bcc', '100', size, a, None, vacuum,
                    periodic=periodic,
                    orthogonal=orthogonal)


def bcc110(symbol, size, a=None, vacuum=None, orthogonal=False,
           periodic=False):
    """BCC(110) surface.

    Supported special adsorption sites: 'ontop', 'longbridge',
    'shortbridge', 'hollow'.

    Use *orthogonal=True* to get an orthogonal unit cell - works only
    for size=(i,j,k) with j even."""
    return _surface(symbol, 'bcc', '110', size, a, None, vacuum,
                    periodic=periodic,
                    orthogonal=orthogonal)


def bcc111(symbol, size, a=None, vacuum=None, orthogonal=False,
           periodic=False):
    """BCC(111) surface.

    Supported special adsorption sites: 'ontop'.

    Use *orthogonal=True* to get an orthogonal unit cell - works only
    for size=(i,j,k) with j even."""
    return _surface(symbol, 'bcc', '111', size, a, None, vacuum,
                    periodic=periodic,
                    orthogonal=orthogonal)


def fcc111(symbol, size, a=None, vacuum=None, orthogonal=False,
           periodic=False):
    """FCC(111) surface.

    Supported special adsorption sites: 'ontop', 'bridge', 'fcc' and 'hcp'.

    Use *orthogonal=True* to get an orthogonal unit cell - works only
    for size=(i,j,k) with j even."""
    return _surface(symbol, 'fcc', '111', size, a, None, vacuum,
                    periodic=periodic,
                    orthogonal=orthogonal)


def hcp0001(symbol, size, a=None, c=None, vacuum=None, orthogonal=False,
            periodic=False):
    """HCP(0001) surface.

    Supported special adsorption sites: 'ontop', 'bridge', 'fcc' and 'hcp'.

    Use *orthogonal=True* to get an orthogonal unit cell - works only
    for size=(i,j,k) with j even."""
    return _surface(symbol, 'hcp', '0001', size, a, c, vacuum,
                    periodic=periodic,
                    orthogonal=orthogonal)


def hcp10m10(symbol, size, a=None, c=None, vacuum=None, orthogonal=True,
             periodic=False):
    """HCP(10m10) surface.

    Supported special adsorption sites: 'ontop'.

    Works only for size=(i,j,k) with j even."""
    if not orthogonal:
        raise NotImplementedError("Can't do non-orthogonal cell yet!")

    return _surface(symbol, 'hcp', '10m10', size, a, c, vacuum,
                    periodic=periodic,
                    orthogonal=orthogonal)


def diamond100(symbol, size, a=None, vacuum=None, orthogonal=True,
               periodic=False):
    """DIAMOND(100) surface.

    Supported special adsorption sites: 'ontop'."""
    if not orthogonal:
        raise NotImplementedError("Can't do non-orthogonal cell yet!")

    return _surface(symbol, 'diamond', '100', size, a, None, vacuum,
                    periodic=periodic,
                    orthogonal=orthogonal)


def diamond111(symbol, size, a=None, vacuum=None, orthogonal=False,
               periodic=False):
    """DIAMOND(111) surface.

    Supported special adsorption sites: 'ontop'."""

    if orthogonal:
        raise NotImplementedError("Can't do orthogonal cell yet!")
    return _surface(symbol, 'diamond', '111', size, a, None, vacuum,
                    periodic=periodic,
                    orthogonal=orthogonal)


def add_adsorbate(slab, adsorbate, height, position=(0, 0), offset=None,
                  mol_index=0):
    """Add an adsorbate to a surface.

    This function adds an adsorbate to a slab.  If the slab is
    produced by one of the utility functions in ase.build, it
    is possible to specify the position of the adsorbate by a keyword
    (the supported keywords depend on which function was used to
    create the slab).

    If the adsorbate is a molecule, the atom indexed by the mol_index
    optional argument is positioned on top of the adsorption position
    on the surface, and it is the responsibility of the user to orient
    the adsorbate in a sensible way.

    This function can be called multiple times to add more than one
    adsorbate.

    Parameters:

    slab: The surface onto which the adsorbate should be added.

    adsorbate:  The adsorbate. Must be one of the following three types:
        A string containing the chemical symbol for a single atom.
        An atom object.
        An atoms object (for a molecular adsorbate).

    height: Height above the surface.

    position: The x-y position of the adsorbate, either as a tuple of
        two numbers or as a keyword (if the surface is produced by one
        of the functions in ase.build).

    offset (default: None): Offsets the adsorbate by a number of unit
        cells. Mostly useful when adding more than one adsorbate.

    mol_index (default: 0): If the adsorbate is a molecule, index of
        the atom to be positioned above the location specified by the
        position argument.

    Note *position* is given in absolute xy coordinates (or as
    a keyword), whereas offset is specified in unit cells.  This
    can be used to give the positions in units of the unit cell by
    using *offset* instead.

    """
    info = slab.info.get('adsorbate_info', {})

    pos = np.array([0.0, 0.0])  # (x, y) part
    spos = np.array([0.0, 0.0])  # part relative to unit cell
    if offset is not None:
        spos += np.asarray(offset, float)

    if isinstance(position, basestring):
        # A site-name:
        if 'sites' not in info:
            raise TypeError('If the atoms are not made by an ' +
                            'ase.build function, ' +
                            'position cannot be a name.')
        if position not in info['sites']:
            raise TypeError('Adsorption site %s not supported.' % position)
        spos += info['sites'][position]
    else:
        pos += position

    if 'cell' in info:
        cell = info['cell']
    else:
        cell = slab.get_cell()[:2, :2]

    pos += np.dot(spos, cell)

    # Convert the adsorbate to an Atoms object
    if isinstance(adsorbate, Atoms):
        ads = adsorbate
    elif isinstance(adsorbate, Atom):
        ads = Atoms([adsorbate])
    else:
        # Assume it is a string representing a single Atom
        ads = Atoms([Atom(adsorbate)])

    # Get the z-coordinate:
    if 'top layer atom index' in info:
        a = info['top layer atom index']
    else:
        a = slab.positions[:, 2].argmax()
        if 'adsorbate_info' not in slab.info:
            slab.info['adsorbate_info'] = {}
        slab.info['adsorbate_info']['top layer atom index'] = a
    z = slab.positions[a, 2] + height

    # Move adsorbate into position
    ads.translate([pos[0], pos[1], z] - ads.positions[mol_index])

    # Attach the adsorbate
    slab.extend(ads)


def add_vacuum(atoms, vacuum):
    """Add vacuum layer to the atoms.

    Parameters:

    atoms: Atoms object
        Most likely created by one of the surface functions.
    vacuum: float
        The thickness of the vacuum layer (in Angstrom).
    """
    uc = atoms.get_cell()
    normal = np.cross(uc[0], uc[1])
    costheta = np.dot(normal, uc[2]) / np.sqrt(np.dot(normal, normal) *
                                               np.dot(uc[2], uc[2]))
    length = np.sqrt(np.dot(uc[2], uc[2]))
    newlength = length + vacuum / costheta
    uc[2] *= newlength / length
    atoms.set_cell(uc)


def _surface(symbol, structure, face, size, a, c, vacuum, periodic,
             orthogonal=True):
    """Function to build often used surfaces.

    Don't call this function directly - use fcc100, fcc110, bcc111, ..."""

    Z = atomic_numbers[symbol]

    if a is None:
        sym = reference_states[Z]['symmetry']
        if sym != structure:
            raise ValueError("Can't guess lattice constant for %s-%s!" %
                             (structure, symbol))
        a = reference_states[Z]['a']

    if structure == 'hcp' and c is None:
        if reference_states[Z]['symmetry'] == 'hcp':
            c = reference_states[Z]['c/a'] * a
        else:
            c = sqrt(8 / 3.0) * a

    positions = np.empty((size[2], size[1], size[0], 3))
    positions[..., 0] = np.arange(size[0]).reshape((1, 1, -1))
    positions[..., 1] = np.arange(size[1]).reshape((1, -1, 1))
    positions[..., 2] = np.arange(size[2]).reshape((-1, 1, 1))

    numbers = np.ones(size[0] * size[1] * size[2], int) * Z

    tags = np.empty((size[2], size[1], size[0]), int)
    tags[:] = np.arange(size[2], 0, -1).reshape((-1, 1, 1))

    slab = Atoms(numbers,
                 tags=tags.ravel(),
                 pbc=(True, True, periodic),
                 cell=size)

    surface_cell = None
    sites = {'ontop': (0, 0)}
    surf = structure + face
    if surf == 'fcc100':
        cell = (sqrt(0.5), sqrt(0.5), 0.5)
        positions[-2::-2, ..., :2] += 0.5
        sites.update({'hollow': (0.5, 0.5), 'bridge': (0.5, 0)})
    elif surf == 'diamond100':
        cell = (sqrt(0.5), sqrt(0.5), 0.5 / 2)
        positions[-4::-4, ..., :2] += (0.5, 0.5)
        positions[-3::-4, ..., :2] += (0.0, 0.5)
        positions[-2::-4, ..., :2] += (0.0, 0.0)
        positions[-1::-4, ..., :2] += (0.5, 0.0)
    elif surf == 'fcc110':
        cell = (1.0, sqrt(0.5), sqrt(0.125))
        positions[-2::-2, ..., :2] += 0.5
        sites.update({'hollow': (0.5, 0.5), 'longbridge': (0.5, 0),
                      'shortbridge': (0, 0.5)})
    elif surf == 'bcc100':
        cell = (1.0, 1.0, 0.5)
        positions[-2::-2, ..., :2] += 0.5
        sites.update({'hollow': (0.5, 0.5), 'bridge': (0.5, 0)})
    else:
        if orthogonal and size[1] % 2 == 1:
            raise ValueError(("Can't make orthorhombic cell with size=%r.  " %
                              (tuple(size),)) +
                             'Second number in size must be even.')
        if surf == 'fcc111':
            cell = (sqrt(0.5), sqrt(0.375), 1 / sqrt(3))
            if orthogonal:
                positions[-1::-3, 1::2, :, 0] += 0.5
                positions[-2::-3, 1::2, :, 0] += 0.5
                positions[-3::-3, 1::2, :, 0] -= 0.5
                positions[-2::-3, ..., :2] += (0.0, 2.0 / 3)
                positions[-3::-3, ..., :2] += (0.5, 1.0 / 3)
            else:
                positions[-2::-3, ..., :2] += (-1.0 / 3, 2.0 / 3)
                positions[-3::-3, ..., :2] += (1.0 / 3, 1.0 / 3)
            sites.update({'bridge': (0.5, 0), 'fcc': (1.0 / 3, 1.0 / 3),
                          'hcp': (2.0 / 3, 2.0 / 3)})
        elif surf == 'diamond111':
            cell = (sqrt(0.5), sqrt(0.375), 1 / sqrt(3) / 2)
            assert not orthogonal
            positions[-1::-6, ..., :3] += (0.0, 0.0, 0.5)
            positions[-2::-6, ..., :2] += (0.0, 0.0)
            positions[-3::-6, ..., :3] += (-1.0 / 3, 2.0 / 3, 0.5)
            positions[-4::-6, ..., :2] += (-1.0 / 3, 2.0 / 3)
            positions[-5::-6, ..., :3] += (1.0 / 3, 1.0 / 3, 0.5)
            positions[-6::-6, ..., :2] += (1.0 / 3, 1.0 / 3)
        elif surf == 'hcp0001':
            cell = (1.0, sqrt(0.75), 0.5 * c / a)
            if orthogonal:
                positions[:, 1::2, :, 0] += 0.5
                positions[-2::-2, ..., :2] += (0.0, 2.0 / 3)
            else:
                positions[-2::-2, ..., :2] += (-1.0 / 3, 2.0 / 3)
            sites.update({'bridge': (0.5, 0), 'fcc': (1.0 / 3, 1.0 / 3),
                          'hcp': (2.0 / 3, 2.0 / 3)})
        elif surf == 'hcp10m10':
            cell = (1.0, 0.5 * c / a, sqrt(0.75))
            assert orthogonal
            positions[-2::-2, ..., 0] += 0.5
            positions[:, ::2, :, 2] += 2.0 / 3
        elif surf == 'bcc110':
            cell = (1.0, sqrt(0.5), sqrt(0.5))
            if orthogonal:
                positions[:, 1::2, :, 0] += 0.5
                positions[-2::-2, ..., :2] += (0.0, 1.0)
            else:
                positions[-2::-2, ..., :2] += (-0.5, 1.0)
            sites.update({'shortbridge': (0, 0.5),
                          'longbridge': (0.5, 0),
                          'hollow': (0.375, 0.25)})
        elif surf == 'bcc111':
            cell = (sqrt(2), sqrt(1.5), sqrt(3) / 6)
            if orthogonal:
                positions[-1::-3, 1::2, :, 0] += 0.5
                positions[-2::-3, 1::2, :, 0] += 0.5
                positions[-3::-3, 1::2, :, 0] -= 0.5
                positions[-2::-3, ..., :2] += (0.0, 2.0 / 3)
                positions[-3::-3, ..., :2] += (0.5, 1.0 / 3)
            else:
                positions[-2::-3, ..., :2] += (-1.0 / 3, 2.0 / 3)
                positions[-3::-3, ..., :2] += (1.0 / 3, 1.0 / 3)
            sites.update({'hollow': (1.0 / 3, 1.0 / 3)})
        else:
            2 / 0

        surface_cell = a * np.array([(cell[0], 0),
                                     (cell[0] / 2, cell[1])])
        if not orthogonal:
            cell = np.array([(cell[0], 0, 0),
                             (cell[0] / 2, cell[1], 0),
                             (0, 0, cell[2])])

    if surface_cell is None:
        surface_cell = a * np.diag(cell[:2])

    if isinstance(cell, tuple):
        cell = np.diag(cell)

    slab.set_positions(positions.reshape((-1, 3)))
    slab.set_cell([a * v * n for v, n in zip(cell, size)], scale_atoms=True)

    if not periodic:
        slab.cell[2] = 0.0

    if vacuum is not None:
        slab.center(vacuum, axis=2)

    if 'adsorbate_info' not in slab.info:
        slab.info.update({'adsorbate_info': {}})

    slab.info['adsorbate_info']['cell'] = surface_cell
    slab.info['adsorbate_info']['sites'] = sites
    return slab


def fcc211(symbol, size, a=None, vacuum=None, orthogonal=True):
    """FCC(211) surface.

    Does not currently support special adsorption sites.

    Currently only implemented for *orthogonal=True* with size specified
    as (i, j, k), where i, j, and k are number of atoms in each direction.
    i must be divisible by 3 to accommodate the step width.
    """
    if not orthogonal:
        raise NotImplementedError('Only implemented for orthogonal '
                                  'unit cells.')
    if size[0] % 3 != 0:
        raise NotImplementedError('First dimension of size must be '
                                  'divisible by 3.')
    atoms = FaceCenteredCubic(symbol,
                              directions=[[1, -1, -1],
                                          [0, 2, -2],
                                          [2, 1, 1]],
                              miller=(None, None, (2, 1, 1)),
                              latticeconstant=a,
                              size=(1, 1, 1),
                              pbc=True)
    z = (size[2] + 1) // 2
    atoms = atoms.repeat((size[0] // 3, size[1], z))
    if size[2] % 2:  # Odd: remove bottom layer and shrink cell.
        remove_list = [atom.index for atom in atoms
                       if atom.z < atoms[1].z]
        del atoms[remove_list]
        dz = atoms[0].z
        atoms.translate((0., 0., -dz))
        atoms.cell[2][2] -= dz

    atoms.cell[2] = 0.0
    atoms.pbc[2] = False
    if vacuum:
        atoms.center(vacuum, axis=2)

    # Renumber systematically from top down.
    orders = [(atom.index, round(atom.x, 3), round(atom.y, 3),
               -round(atom.z, 3), atom.index) for atom in atoms]
    orders.sort(key=itemgetter(3, 1, 2))
    newatoms = atoms.copy()
    for index, order in enumerate(orders):
        newatoms[index].position = atoms[order[0]].position.copy()

    # Add empty 'sites' dictionary for consistency with other functions
    newatoms.info['adsorbate_info'] = {'sites': {}}
    return newatoms


def mx2(formula='MoS2', kind='2H', a=3.18, thickness=3.19,
        size=(1, 1, 1), vacuum=None):
    """Create three-layer 2D materials with hexagonal structure.

    For metal dichalcogenites, etc.

    The kind argument accepts '2H', which gives a mirror plane symmetry
    and '1T', which gives an inversion symmetry."""

    if kind == '2H':
        basis = [(0, 0, 0),
                 (2 / 3, 1 / 3, 0.5 * thickness),
                 (2 / 3, 1 / 3, -0.5 * thickness)]
    elif kind == '1T':
        basis = [(0, 0, 0),
                 (2 / 3, 1 / 3, 0.5 * thickness),
                 (1 / 3, 2 / 3, -0.5 * thickness)]
    else:
        raise ValueError('Structure not recognized:', kind)

    cell = [[a, 0, 0], [-a / 2, a * 3**0.5 / 2, 0], [0, 0, 0]]

    atoms = Atoms(formula, cell=cell, pbc=(1, 1, 0))
    atoms.set_scaled_positions(basis)
    if vacuum is not None:
        atoms.center(vacuum, axis=2)
    atoms = atoms.repeat(size)
    return atoms


def graphene(formula='C2', a=2.460, size=(1, 1, 1), vacuum=None):
    """Create a graphene monolayer structure."""
    cell = [[a, 0, 0], [-a / 2, a * 3**0.5 / 2, 0], [0, 0, 0]]
    basis = [[0, 0, 0], [2./3, 1./3, 0]]
    atoms = Atoms(formula, cell=cell, pbc=(1, 1, 0))
    atoms.set_scaled_positions(basis)
    if vacuum is not None:
        atoms.center(vacuum, axis=2)
    atoms = atoms.repeat(size)
    return atoms


def _all_surface_functions():
    # Convenient for debugging.
    d = {}
    for func in [fcc100, fcc110, bcc100, bcc110, bcc111, fcc111, hcp0001,
                 hcp10m10, diamond100, diamond111, fcc111, mx2, graphene]:
        d[func.__name__] = func
    return d



import numpy as np

from ase.data import chemical_symbols
from ase.atoms import Atoms
from ase.utils import basestring


iwm_symbols = {'1': 'C',
               '2': 'Au',
               '5': 'Ag'}


def read_iwm(fileobj, index=-1):
    if isinstance(fileobj, basestring):
        fileobj = open(fileobj)

    lines = fileobj.readlines()
    L1 = lines[1].split()
    if len(L1) == 1:
        del lines[:3]
        natoms = int(L1[0])
    else:
        natoms = len(lines)
    images = []

    positions = []
    symbols = []
    for line in lines[:natoms]:
        symbol, mass, x, y, z = line.split()[:5]
        if symbol in iwm_symbols:
            symbols.append(iwm_symbols[symbol])
        else:
            symbols.append(chemical_symbols[int(symbol)])
        positions.append([float(x), float(y), float(z)])
    
    del(lines[natoms:3 * natoms + 3])
    
    cell = []
    for line in lines[natoms:natoms + 3]:
        x, y, z = line.split()[:3]
        cell.append(np.array([float(x), float(y), float(z)]))
      
    images.append(Atoms(symbols=symbols, positions=positions, cell=cell))

    return images[index]



# creates: Au-wire.png

from ase import Atoms
from ase.io import write

d = 2.9
L = 10.0
wire = Atoms('Au',
             positions=[(0, L / 2, L / 2)],
             cell=(d, L, L),
             pbc=(1, 0, 0))
wire *= (6, 1, 1)
wire.positions[:, 0] -= 2 * d
wire.cell[0, 0] = d
#view(wire, block=1)
write('Au-wire.pov', wire,
      rotation='12x,6y',
      transparent=False, run_povray=True)



# flake8: noqa
'''
mayavi interface to plot atoms, unit cells, and volumetric data
'''
import numpy as np
from enthought.mayavi import mlab
mlab.figure(1, bgcolor=(1,1,1), size=(350, 350))
mlab.clf()

def plot_cylinder(start,end,tube_radius=0.1,color=(0,0,0)):

    mlab.plot3d([start[0],end[0]],[start[1],end[1]],[start[2],end[2]],
                tube_radius=tube_radius,color=color)


def plot_atoms(atoms):

    for atom in atoms:
        pos = atom.position
        mlab.points3d([pos[0]],[pos[1]],[pos[2]],
                      scale_factor=4,
                      resolution=20,
                      color=(1,0,0),        #this should get species specifuc
                      scale_mode='none')

    (u0,u1,u2) = atoms.get_cell()
    origin = np.array([0.0,0.0,0.0])

    plot_cylinder(origin,u0)
    plot_cylinder(origin,u1)
    plot_cylinder(origin,u2)

    plot_cylinder(u0,u0+u1)
    plot_cylinder(u0,u0+u2)

    plot_cylinder(u1,u1+u0)
    plot_cylinder(u1,u1+u2)

    plot_cylinder(u2,u2+u0)
    plot_cylinder(u2,u2+u1)

    plot_cylinder(u0+u1,u0+u1+u2)
    plot_cylinder(u1+u2,u0+u1+u2)
    plot_cylinder(u0+u2,u0+u1+u2)
    mlab.show()






if __name__ == '__main__':
    from ase.lattice.cubic import FaceCenteredCubic

    from ase.lattice.bravais import cross

    a = np.array([0.5,0,0])
    c = np.array([0,1,0],dtype=np.float)
    b1 = c - a

    a = np.array([0,1,0],np.float)
    c = np.array([0,0.5,0.5])
    b2 = c - a

    a3 = np.array([2,1,1],np.float)

    a1 = cross(b1,a3)
    a2 = cross(b2,a3)
    v211 = FaceCenteredCubic(directions=[a1,a2,a3],
                             miller=(None,None,[2,1,1]),
                             symbol='Pd',
                             size=(1,1,2),
                             debug=0)

    uc = v211.get_cell()
    uc[2][2] += 10.0
    v211.set_cell(uc)

    plot_atoms(v211.repeat((2,2,1)))




# -*- coding: utf-8 -*-
import warnings

import numpy as np
from numpy.linalg import eigh

from ase.optimize.optimize import Optimizer
from ase.utils import basestring


class BFGS(Optimizer):
    def __init__(self, atoms, restart=None, logfile='-', trajectory=None,
                 maxstep=0.04, master=None):
        """BFGS optimizer.

        Parameters:

        atoms: Atoms object
            The Atoms object to relax.

        restart: string
            Pickle file used to store hessian matrix. If set, file with
            such a name will be searched and hessian matrix stored will
            be used, if the file exists.

        trajectory: string
            Pickle file used to store trajectory of atomic movement.

        logfile: file object or str
            If *logfile* is a string, a file with that name will be opened.
            Use '-' for stdout.

        maxstep: float
            Used to set the maximum distance an atom can move per
            iteration (default value is 0.04 Å).

        master: boolean
            Defaults to None, which causes only rank 0 to save files.  If
            set to true,  this rank will save files.
        """
        if maxstep > 1.0:
            warnings.warn('You are using a much too large value for '
                          'the maximum step size: %.1f Å' % maxstep)
        self.maxstep = maxstep

        Optimizer.__init__(self, atoms, restart, logfile, trajectory, master)

    def todict(self):
        d = Optimizer.todict(self)
        if hasattr(self, 'maxstep'):
            d.update(maxstep=self.maxstep)
        return d

    def initialize(self):
        self.H = None
        self.r0 = None
        self.f0 = None

    def read(self):
        self.H, self.r0, self.f0, self.maxstep = self.load()

    def step(self, f=None):
        atoms = self.atoms

        if f is None:
            f = atoms.get_forces()

        r = atoms.get_positions()
        f = f.reshape(-1)
        self.update(r.flat, f, self.r0, self.f0)
        omega, V = eigh(self.H)
        dr = np.dot(V, np.dot(f, V) / np.fabs(omega)).reshape((-1, 3))
        steplengths = (dr**2).sum(1)**0.5
        dr = self.determine_step(dr, steplengths)
        atoms.set_positions(r + dr)
        self.r0 = r.flat.copy()
        self.f0 = f.copy()
        self.dump((self.H, self.r0, self.f0, self.maxstep))

    def determine_step(self, dr, steplengths):
        """Determine step to take according to maxstep

        Normalize all steps as the largest step. This way
        we still move along the eigendirection.
        """
        maxsteplength = np.max(steplengths)
        if maxsteplength >= self.maxstep:
            dr *= self.maxstep / maxsteplength

        return dr

    def update(self, r, f, r0, f0):
        if self.H is None:
            self.H = np.eye(3 * len(self.atoms)) * 70.0
            return
        dr = r - r0

        if np.abs(dr).max() < 1e-7:
            # Same configuration again (maybe a restart):
            return

        df = f - f0
        a = np.dot(dr, df)
        dg = np.dot(self.H, dr)
        b = np.dot(dr, dg)
        self.H -= np.outer(df, df) / a + np.outer(dg, dg) / b

    def replay_trajectory(self, traj):
        """Initialize hessian from old trajectory."""
        if isinstance(traj, basestring):
            from ase.io.trajectory import Trajectory
            traj = Trajectory(traj, 'r')
        self.H = None
        atoms = traj[0]
        r0 = atoms.get_positions().ravel()
        f0 = atoms.get_forces().ravel()
        for atoms in traj:
            r = atoms.get_positions().ravel()
            f = atoms.get_forces().ravel()
            self.update(r, f, r0, f0)
            r0 = r
            f0 = f

        self.r0 = r0
        self.f0 = f0


class oldBFGS(BFGS):
    def determine_step(self, dr, steplengths):
        """Old BFGS behaviour for scaling step lengths

        This keeps the behaviour of truncating individual steps. Some might
        depend of this as some absurd kind of stimulated annealing to find the
        global minimum.
        """
        dr /= np.maximum(steplengths / self.maxstep, 1.0).reshape(-1, 1)
        return dr



import numpy as np
import scipy  # skip test early if no scipy
from ase.build import bulk
from ase.calculators.emt import EMT
from ase.eos import EquationOfState as EOS, eos_names
scipy  # silence pyflakes

b = bulk('Al', 'fcc', a=4.0, orthorhombic=True)
b.set_calculator(EMT())
cell = b.get_cell()

volumes = []
energies = []
for x in np.linspace(0.98, 1.01, 5):
    b.set_cell(cell * x, scale_atoms=True)
    volumes.append(b.get_volume())
    energies.append(b.get_potential_energy())

results = []
for name in eos_names:
    if name == 'antonschmidt':
        # Someone should fix this!
        continue
    eos = EOS(volumes, energies, name)
    v, e, b = eos.fit()
    print('{0:20} {1:.8f} {2:.8f} {3:.8f} '.format(name, v, e, b))
    assert abs(v - 3.18658700e+01) < 4e-4
    assert abs(e - -9.76187802e-03) < 5e-7
    assert abs(b - 2.46812688e-01) < 2e-4
    results.append((v, e, b))

print(np.ptp(results, 0))
print(np.mean(results, 0))



from ase.utils import basestring

# Copyright (C) 2010 by CAMd, DTU
# Please see the accompanying LICENSE file for further information.

# This file is taken (almost) verbatim from CMR with D. Landis agreement

FIELD_SEPARATOR = "\\"
PARA_START = "\n\n"
PARA_END = "\\\\@"

names = ['', '', 'Computer_system', 'Type_of_run', 'Method', 'Basis_set',
         'Chemical_formula', 'Person', 'Date', '', '', '', '', 'Title', '']
names_compact = ['', '', 'Computer_system', 'Type_of_run', 'Method',
                 'Basis_set', 'Chemical_formula', 'Person', 'Date', '', '', '',
                 '', 'Title', '']

charge_multiplicity = 15


class GaussianReader:

    def auto_type(self, data):
        """ tries to determine type"""
        try:
            return float(data)
        except ValueError:
            pass

        try:
            ds = data.split(",")
            array = []

            for d in ds:
                array.append(float(d))

            return array
        except ValueError:
            pass

        return data

    def __init__(self, filename, read_structures=False):
        """filename is NOT optional"""
        if isinstance(filename, basestring):
            fileobj = open(filename, 'r')
        elif hasattr(filename,'seek'):
            fileobj = filename
            fileobj.seek(0)  # Re-wind fileobj
        else:
            msg = 'Cannot use given filename, make sure it is a string or a fileobject'
            raise RuntimeError(msg)

        content = fileobj.read()

# handles the case that users used windows after the calculation:
        content = content.replace("\r\n", "\n")

        self.parse(content)

        #read structures from file
        if read_structures:
            self.read_structures(content)


    def get_structures(self, content=None):
        """Get Structures"""
        if hasattr(self,'structures'):
            return self.structures
        elif content is None:
            raise RuntimeError('Images not available and no content parsed!')
        else:
            self.read_structures(content)
            return self.structures


    def read_structures(self, content=None):
        """Read Structures from file and wirte them to self.structures"""
        from ase.atoms import Atoms
        from ase.atom import Atom
        images = []
        temp_items = content.split('Standard orientation')[1:]
        for item_i in temp_items:
            lines = [ line for line in item_i.split('\n') if len(line) > 0 ]
            #first 5 lines are headers
            del lines[:5]
            images.append(Atoms())
            for line in lines:
                #if only - in line it is the end
                if set(line).issubset(set('- ')):
                    break
                tmp_line = line.strip().split()
                if not len(tmp_line) == 6:
                    raise RuntimeError('Length of line does not match structure!')

                #read atom
                try:
                    atN = int(tmp_line[1])
                    pos = tuple(float(x) for x in tmp_line[3:])
                except ValueError:
                    raise ValueError('Expected a line with three integers and three floats.')
                images[-1].append(Atom(atN,pos))
        self.structures = images
        return



    def parse(self, content):
        from ase.data import atomic_numbers
        chg_mult = charge_multiplicity
        self.data = []
        temp_items = content.split(PARA_START)
        seq_count = 0
        for i in temp_items:
            i = i.replace("\n ", "")
            if i.endswith(PARA_END):
                i = i.replace(PARA_END, "")
                i = i.split(FIELD_SEPARATOR)

                new_dict = {}
                self.data.append(new_dict)

                new_dict['Sequence number'] = seq_count
                seq_count += 1
                for pos in range(len(names)):
                    if names[pos] != "":
                        #hack, since this section is too short if there is no title
                        if names[pos] == "Title" and i[pos] == "":
                            chg_mult -= 1
                            break
                        new_dict[names[pos]] = self.auto_type(i[pos])

                chm = i[chg_mult].split(",")
                new_dict["Charge"] = int(chm[0])
                new_dict["Multiplicity"] = int(chm[1])

# Read atoms
                atoms = []
                positions = []
                position = chg_mult + 1
                while position < len(i) and i[position] != "":
                    s = i[position].split(",")
                    atoms.append(atomic_numbers[s[0].capitalize()])
                    #if fragments are specified, there are 4 numbers
                    #first one integer and then xyz coords
                    #therefore use xyz from the end
                    positions.append([float(s[-3]), float(s[-2]), float(s[-1])])
                    position = position + 1

                new_dict["Atomic_numbers"] = atoms
                new_dict["Positions"] = positions
# Read more variables
                position += 1
                while position < len(i) and i[position] != "":
                    s = i[position].split('=')
                    if len(s) == 2:
                        new_dict[s[0]] = self.auto_type(s[1])
                    else:
                        print("Warning: unexpected input ", s)
                    position = position + 1

    def __iter__(self):
        """returns an iterator that iterates over all keywords"""
        return self.data.__iter__()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, pos):
        return self.data[pos]



"""Test run for DFTB+ calculator.

Use these .skf-files from dftb.org:

    * trans3d/trans3d-0-1/Ni-Ni.skf
    * trans3d/trans3d-0-1/Ni-N.skf
    * mio/mio-1-1/N-N.skf

"""
import os

from ase import Atoms
from ase.calculators.dftb import Dftb
from ase.constraints import FixAtoms
from ase.optimize import QuasiNewton
from ase.build import fcc111, add_adsorbate

h = 1.85
d = 1.10
k = 2

slab = fcc111('Ni', size=(2, 2, 3), vacuum=10.0)
calc1 = Dftb(label='slab',
             kpts=[k, k, 1],
             Hamiltonian_SCC='YES')
slab.set_calculator(calc1)
dyn = QuasiNewton(slab, trajectory='slab.traj')
dyn.run(fmax=0.05)
e_slab = slab.get_potential_energy()
os.system('rm dftb_in.hsd')
molecule = Atoms('2N', positions=[(0., 0., 0.), (0., 0., d)])
calc2 = Dftb(label='n2',
             Hamiltonian_SCC='YES')
molecule.set_calculator(calc2)
dyn = QuasiNewton(molecule, trajectory='n2.traj')
dyn.run(fmax=0.05)
e_N2 = molecule.get_potential_energy()

slab2 = slab
add_adsorbate(slab2, molecule, h, 'ontop')
constraint = FixAtoms(mask=[a.symbol != 'N' for a in slab2])
slab2.set_constraint(constraint)
calc3 = Dftb(label='slab2',
             kpts=[k, k, 1],
             Hamiltonian_SCC='YES')
slab2.set_calculator(calc3)
dyn = QuasiNewton(slab2, trajectory='slab2.traj')
dyn.run(fmax=0.05)

adsorption_energy = e_slab + e_N2 - slab2.get_potential_energy()
print(adsorption_energy, 'eV')
assert abs(adsorption_energy - -0.30) < 0.1



# flake8: noqa
"""An experimental package for making plots during a simulation.

A PrimiPlotter can plot a list of atoms on one or more output devices.
"""

import collections
import os
import sys
import time
import weakref
from math import pi

import numpy as np

from ase.visualize.colortable import color_table
import ase.data
from ase.utils import basestring


class PrimiPlotterBase:
    "Base class for PrimiPlotter and Povrayplotter."
    #def set_dimensions(self, dims):
    #    "Set the size of the canvas (a 2-tuple)."
    #    self.dims = dims

    def set_rotation(self, rotation):
        "Set the rotation angles (in degrees)."
        self.angles[:] = np.array(rotation) * (pi/180)

    def set_radii(self, radii):
        """Set the atomic radii.  Give an array or a single number."""
        self.radius = radii

    def set_colors(self, colors):
        """Explicitly set the colors of the atoms.
        
        The colors can either be a dictionary mapping tags to colors
        or an array of colors, one per atom.
        
        Each color is specified as a greyscale value from 0.0 to 1.0
        or as three RGB values from 0.0 to 1.0.
        """
        self.colors = colors

    def set_color_function(self, colors):
        """Set a color function, to be used to color the atoms."""
        if isinstance(colors, collections.Callable):
            self.colorfunction = colors
        else:
            raise TypeError("The color function is not callable.")

    def set_invisible(self, inv):
        """Choose invisible atoms."""
        self.invisible = inv

    def set_invisibility_function(self, invfunc):
        """Set an invisibility function."""
        if isinstance(invfunc, collections.Callable):
            self.invisibilityfunction = invfunc
        else:
            raise TypeError("The invisibility function is not callable.")

    def set_cut(self, xmin=None, xmax=None, ymin=None, ymax=None,
                zmin=None, zmax=None):
        self.cut = {"xmin":xmin, "xmax":xmax, "ymin":ymin, "ymax":ymax,
                    "zmin":zmin, "zmax":zmax}
    
    def update(self, newatoms = None):
        """Cause a plot (respecting the interval setting).

        update causes a plot to be made.  If the interval variable was
        specified when the plotter was create, it will only produce a
        plot with that interval.  update takes an optional argument,
        newatoms, which can be used to replace the list of atoms with
        a new one.
        """
        if newatoms is not None:
            self.atoms = newatoms
        if self.skipnext <= 0:
            self.plot()
            self.skipnext = self.interval
        self.skipnext -= 1
        
    def set_log(self, log):
        """Sets a file for logging.

        log may be an open file or a filename.
        """
        if hasattr(log, "write"):
            self.logfile = log
            self.ownlogfile = False
        else:
            self.logfile = open(log, "w")
            self.ownlogfile = True

    def log(self, message):
        """logs a message to the file set by set_log."""
        if self.logfile is not None:
            self.logfile.write(message+"\n")
            self.logfile.flush()
        self._verb(message)
        
    def _verb(self, txt):
        if self.verbose:
            sys.stderr.write(txt+"\n")
    
    def _starttimer(self):
        self.starttime = time.time()

    def _stoptimer(self):
        elapsedtime = time.time() - self.starttime
        self.totaltime = self.totaltime + elapsedtime
        print("plotting time %s sec (total %s sec)" % (elapsedtime,
                                                       self.totaltime))

    def _getpositions(self):
        return self.atoms.get_positions()

    def _getradii(self):
        if self.radius is not None:
            if hasattr(self.radius, "shape"):
                return self.radius   # User has specified an array
            else:
                return self.radius * np.ones(len(self.atoms), float)
        # No radii specified.  Try getting them from the atoms.
        try:
            return self.atoms.get_atomic_radii()
        except AttributeError:
            try:
                z = self._getatomicnumbers()
            except AttributeError:
                pass
            else:
                return ase.data.covalent_radii[z]
        # No radius available.  Defaulting to 1.0
        return np.ones(len(self.atoms), float)

    def _getatomicnumbers(self):
        return self.atoms.get_atomic_numbers()
    
    def _getcolors(self):
        # Try any explicitly given colors
        if self.colors is not None:
            if isinstance(self.colors, type({})):
                self.log("Explicit colors dictionary")
                return _colorsfromdict(self.colors,
                                       np.asarray(self.atoms.get_tags(),int))
            else:
                self.log("Explicit colors")
                return self.colors
        # Try the color function, if given
        if self.colorfunction is not None:
            self.log("Calling color function.")
            return self.colorfunction(self.atoms)
        # Maybe the atoms know their own colors
        try:
            c = self.atoms.get_colors()
        except AttributeError:
            c = None
        if c is not None:
            if isinstance(c, type({})):
                self.log("Color dictionary from atoms.get_colors()")
                return _colorsfromdict(c, np.asarray(self.atoms.get_tags(),int))
            else:
                self.log("Colors from atoms.get_colors()")
                return c
        # Default to white atoms
        self.log("No colors: using white")
        return np.ones(len(self.atoms), float)

    def _getinvisible(self):
        if self.invisible is not None:
            inv = self.invisible
        else:
            inv = np.zeros(len(self.atoms))
        if self.invisibilityfunction:
            inv = np.logical_or(inv, self.invisibilityfunction(self.atoms))
        r = self._getpositions()
        if len(r) > len(inv):
            # This will happen in parallel simulations due to ghost atoms.
            # They are invisible.  Hmm, this may cause trouble.
            i2 = np.ones(len(r))
            i2[:len(inv)] = inv
            inv = i2
            del i2
        if self.cut["xmin"] is not None:
            inv = np.logical_or(inv, np.less(r[:,0], self.cut["xmin"]))
        if self.cut["xmax"] is not None:
            inv = np.logical_or(inv, np.greater(r[:,0], self.cut["xmax"]))
        if self.cut["ymin"] is not None:
            inv = np.logical_or(inv, np.less(r[:,1], self.cut["ymin"]))
        if self.cut["ymax"] is not None:
            inv = np.logical_or(inv, np.greater(r[:,1], self.cut["ymax"]))
        if self.cut["zmin"] is not None:
            inv = np.logical_or(inv, np.less(r[:,2], self.cut["zmin"]))
        if self.cut["zmax"] is not None:
            inv = np.logical_or(inv, np.greater(r[:,2], self.cut["zmax"]))
        return inv

    def __del__(self):
        if self.ownlogfile:
            self.logfile.close()
            
class PrimiPlotter(PrimiPlotterBase):
    """Primitive PostScript-based plots during a simulation.

    The PrimiPlotter plots atoms during simulations, extracting the
    relevant information from the list of atoms.  It is created using
    the list of atoms as an argument to the constructor.  Then one or
    more output devices must be attached using set_output(device).  The
    list of supported output devices is at the end.

    The atoms are plotted as circles.  The system is first rotated
    using the angles specified by set_rotation([vx, vy, vz]).  The
    rotation is vx degrees around the x axis (positive from the y
    toward the z axis), then vy degrees around the y axis (from x
    toward z), then vz degrees around the z axis (from x toward y).
    The rotation matrix is the same as the one used by RasMol.

    Per default, the system is scaled so it fits within the canvas
    (autoscale mode).  Autoscale mode is enabled and disables using
    autoscale("on") or autoscale("off").  A manual scale factor can be
    set with set_scale(scale), this implies autoscale("off").  The
    scale factor (from the last autoscale event or from set_scale) can
    be obtained with get_scale().  Finally, an explicit autoscaling can
    be triggered with autoscale("now"), this is mainly useful before
    calling get_scale or before disabling further autoscaling.
    Finally, a relative scaling factor can be set with
    SetRelativeScaling(), it is multiplied to the usual scale factor
    (from autoscale or from set_scale).  This is probably only useful in
    connection with autoscaling.

    The radii of the atoms are obtained from the first of the following
    methods which work:
    
    1.  If the radii are specified using PrimiPlotter.set_radii(r),
        they are used.  Must be an array, or a single number.

    2.  If the atoms has a get_atomic_radii() method, it is used.  This is
        unlikely.

    3.  If the atoms has a get_atomic_numbers() method, the
        corresponding covalent radii are extracted from the
        ASE.ChemicalElements module.

    4.  If all else fails, the radius is set to 1.0 Angstrom.

    The atoms are colored using the first of the following methods
    which work.

    1.  If colors are explicitly set using PrimiPlotter.set_colors(),
        they are used.

    2.  If these colors are specified as a dictionary, the tags
        (from atoms.get_tags()) are used as an index into the
        dictionary to get the actual colors of the atoms.

    3.  If a color function has been set using
        PrimiPlotter.set_color_function(), it is called with the atoms
        as an argument, and is expected to return an array of colors.

    4.  If the atoms have a get_colors() method, it is used to get the
        colors.

    5.  If these colors are specified as a dictionary, the tags
        (from atoms.get_tags()) are used as an index into the
        dictionary to get the actual colors of the atoms.

    6.  If all else fails, the atoms will be white.

    The colors are specified as an array of colors, one color per
    atom.  Each color is either a real number from 0.0 to 1.0,
    specifying a grayscale (0.0 = black, 1.0 = white), or an array of
    three numbers from 0.0 to 1.0, specifying RGB values.  The colors
    of all atoms are thus a Numerical Python N-vector or a 3xN matrix.

    In cases 1a and 3a above, the keys of the dictionary are integers,
    and the values are either numbers (grayscales) or 3-vectors (RGB
    values), or strings with X11 color names, which are then
    translated to RGB values.  Only in case 1a and 3a are strings
    recognized as colors.

    Some atoms may be invisible, and thus left out of the plot.
    Invisible atoms are determined from the following algorithm.
    Unlike the radius or the coloring, all points below are tried and
    if an atom is invisible by any criterion, it is left out of the plot.

    1.  All atoms are visible.
    
    2.  If PrimiPlotter.set_invisible() has be used to specify invisible
        atoms, any atoms for which the value is non-zero becomes invisible.

    3.  If an invisiblility function has been set with
        PrimiPlotter.set_invisibility_function(), it is called with the
        atoms as argument.  It is expected to return an integer per
        atom, any non-zero value makes that atom invisible.

    4.  If a cut has been specified using set_cut, any atom outside the
        cut is made invisible.

    Note that invisible atoms are still included in the algorithm for
    positioning and scaling the plot.

    
    The following output devices are implemented.
    
    PostScriptFile(prefix):  Create PS files names prefix0000.ps etc.

    PnmFile(prefix):  Similar, but makes PNM files.

    GifFile(prefix):  Similar, but makes GIF files.

    JpegFile(prefix):  Similar, but makes JPEG files.

    X11Window():  Show the plot in an X11 window using ghostscript.

    Output devices writing to files take an extra optional argument to
    the constructor, compress, specifying if the output file should be
    gzipped.  This is not allowed for some (already compressed) file
    formats.

    Instead of a filename prefix, a filename containing a % can be
    used.  In that case the filename is expected to expand to a real
    filename when used with the Python string formatting operator (%)
    with the frame number as argument.  Avoid generating spaces in the
    file names: use e.g. %03d instead of %3d.
    """
    def __init__(self, atoms, verbose=0, timing=0, interval=1, initframe=0):
        """

        Parameters to the constructor:

        atoms: The atoms to be plottet.

        verbose = 0:  Write progress information to stderr.

        timing = 0:  Collect timing information.

        interval = 1: If specified, a plot is only made every
        interval'th time update() is called.  Deprecated, normally you
        should use the interval argument when attaching the plotter to
        e.g. the dynamics.

        initframe = 0: Initial frame number, i.e. the number of the
        first plot.
        
        """
        self.atoms = atoms
        self.outputdevice = []
        self.angles = np.zeros(3, float)
        self.dims = (512, 512)
        self.verbose = verbose
        self.timing = timing
        self.totaltime = 0.0
        self.radius = None
        self.colors = None
        self.colorfunction = None
        self.n = initframe
        self.interval = interval
        self.skipnext = 0 # Number of calls to update before anything happens.
        self.a_scale = 1
        self.relativescale = 1.0
        self.invisible = None
        self.invisibilityfunction = None
        self.set_cut()   # No cut
        self.isparallel = 0
        self.logfile = None
        self.ownlogfile = False
        
    def set_output(self, device):
        "Attach an output device to the plotter."
        self.outputdevice.append(device)
        device.set_dimensions(self.dims)
        device.set_owner(weakref.proxy(self))

    def set_dimensions(self, dims):
        "Set the size of the canvas (a 2-tuple)."
        if self.outputdevice:
            raise RuntimeError("Cannot set dimensions after an output device has been specified.")
        self.dims = dims
        
    def autoscale(self, mode):
        if mode == "on":
            self.a_scale = 1
        elif mode == "off":
            self.a_scale = 0
        elif mode == "now":
            coords = self._rotate(self.atoms.get_positions())
            radii = self._getradii()
            self._autoscale(coords, radii)
        else:
            raise ValueError("Unknown autoscale mode: ").with_traceback(+str(mode))

    def set_scale(self, scale):
        self.autoscale("off")
        self.scale = scale

    def get_scale(self):
        return self.scale

    def set_relative_scale(self, rscale = 1.0):
        self.relativescale = rscale

    def plot(self):
        """Create a plot now.  Does not respect the interval timer.

        This method makes a plot unconditionally.  It does not look at
        the interval variable, nor is this plot taken into account in
        the counting done by the update() method if an interval
        variable was specified.
        """
        if self.timing:
            self._starttimer()
        self.log("PrimiPlotter: Starting plot at "
                 + time.strftime("%a, %d %b %Y %H:%M:%S"))
        colors = self._getcolors()
        invisible = self._getinvisible()
        coords = self._rotate(self._getpositions())
        radii = self._getradii()
        if self.a_scale:
            self._autoscale(coords,radii)
        scale = self.scale * self.relativescale
        coords = scale * coords
        center = self._getcenter(coords)
        offset = np.array(self.dims + (0.0,))/2.0 - center
        coords = coords + offset
        self.log("Scale is %f and size is (%d, %d)"
                 % (scale, self.dims[0], self.dims[1]))
        self.log("Physical size of plot is %f Angstrom times %f Angstrom"
                 % (self.dims[0] / scale, self.dims[1] / scale))

        self._verb("Sorting.")
        order = np.argsort(coords[:,2])
        coords = coords[order]  ### take(coords, order)
        radii = radii[order]    ### take(radii, order)
        colors = colors[order]  ### take(colors, order)
        invisible = invisible[order]  ### take(invisible, order)
        if self.isparallel:
            id = np.arange(len(coords))[order] ### take(arange(len(coords)), order)
        else:
            id = None
            
        radii = radii * scale
        selector = self._computevisibility(coords, radii, invisible, id)
        coords = np.compress(selector, coords, 0)
        radii = np.compress(selector, radii)
        colors = np.compress(selector, colors, 0)
        self._makeoutput(scale, coords, radii, colors)
        self.log("PrimiPlotter: Finished plotting at "
                 + time.strftime("%a, %d %b %Y %H:%M:%S"))
        self.log("\n\n")
        if self.timing:
            self._stoptimer()

    def _computevisibility(self, coords, rad, invisible, id, zoom = 1):
        xy = coords[:,:2]
        typradius = sum(rad) / len(rad)
        if typradius < 4.0:
            self.log("Refining visibility check.")
            if zoom >= 16:
                raise RuntimeError("Cannot check visibility - too deep recursion.")
            return self._computevisibility(xy*2, rad*2, invisible, id, zoom*2)
        else:
            self.log("Visibility(r_typ = %.1f pixels)" % (typradius,))
        dims = np.array(self.dims) * zoom
        maxr = int(np.ceil(max(rad))) + 2
        canvas = np.zeros((dims[0] + 4*maxr, dims[1] + 4*maxr), np.int8)
        # Atoms are only invisible if they are within the canvas, or closer
        # to its edge than their radius
        visible = (np.greater(xy[:,0], -rad) * np.less(xy[:,0], dims[0]+rad)
                   * np.greater(xy[:,1], -rad) * np.less(xy[:,1], dims[1]+rad)
                   * np.logical_not(invisible))
        # Atoms are visible if not hidden behind other atoms
        xy = np.floor(xy + 2*maxr + 0.5).astype(int)
        masks = {}
        for i in range(len(rad)-1, -1, -1):
            if (i % 100000) == 0 and i:
                self._verb(str(i))
            if not visible[i]:
                continue
            x, y = xy[i]
            r = rad[i]
            try:
                mask, invmask, rn = masks[r]
            except KeyError:
                rn = int(np.ceil(r))
                nmask = 2*rn+1
                mask = (np.arange(nmask) - rn)**2
                mask = np.less(mask[:,np.newaxis]+mask[np.newaxis,:], r*r).astype(np.int8)
                invmask = np.equal(mask, 0).astype(np.int8)
                masks[r] = (mask, invmask, rn)
            window = np.logical_or(canvas[x-rn:x+rn+1, y-rn:y+rn+1], invmask)
            hidden = np.alltrue(window.flat)
            if hidden:
                visible[i] = 0
            else:
                canvas[x-rn:x+rn+1, y-rn:y+rn+1] = np.logical_or(canvas[x-rn:x+rn+1, y-rn:y+rn+1], mask)
        self.log("%d visible, %d hidden out of %d" %
                   (sum(visible), len(visible) - sum(visible), len(visible)))
        return visible
        
    def _rotate(self, positions):
        self.log("Rotation angles: %f %f %f" % tuple(self.angles))
        mat = np.dot(np.dot(_rot(self.angles[2], 2),
                            _rot(self.angles[1], 1)),
                     _rot(self.angles[0]+pi, 0))
        return np.dot(positions, mat)

    def _getcenter(self, coords):
        return np.array((max(coords[:,0]) + min(coords[:,0]),
                      max(coords[:,1]) + min(coords[:,1]), 0.0)) / 2.0

    def _autoscale(self, coords, radii):
        x = coords[:,0]
        y = coords[:,1]
        maxradius = max(radii)
        deltax = max(x) - min(x) + 2*maxradius
        deltay = max(y) - min(y) + 2*maxradius
        scalex = self.dims[0] / deltax
        scaley = self.dims[1] / deltay
        self.scale = 0.95 * min(scalex, scaley)
        self.log("Autoscale: %f" % self.scale)

    def _makeoutput(self, scale, coords, radii, colors):
        for device in self.outputdevice:
            device.inform_about_scale(scale)
            device.plot(self.n, coords, radii, colors)
        self.n = self.n + 1


class ParallelPrimiPlotter(PrimiPlotter):
    """A version of PrimiPlotter for parallel ASAP simulations.

    Used like PrimiPlotter, but only the output devices on the master
    node are used.  Most of the processing is distributed on the
    nodes, but the actual output is only done on the master.  See the
    PrimiPlotter docstring for details.
    """
    def __init__(self, *args, **kwargs):
        PrimiPlotter.__init__(self, *args, **kwargs)
        self.isparallel = 1
        import ase.parallel
        self.mpi = ase.parallel.world
        if self.mpi is None:
            raise RuntimeError("MPI is not available.")
        self.master = self.mpi.rank == 0
        self.mpitag = 42   # Reduce chance of collision with other modules.
        
    def set_output(self, device):
        if self.master:
            PrimiPlotter.set_output(self, device)

    def set_log(self, log):
        if self.master:
            PrimiPlotter.set_log(self, log)

    def _getpositions(self):
        realpos = self.atoms.get_positions()
        ghostpos = self.atoms.get_ghost_positions()
        self.numberofrealatoms = len(realpos)
        self.numberofghostatoms = len(ghostpos)
        return np.concatenate((realpos, ghostpos))

    def _getatomicnumbers(self):
        realz = self.atoms.get_atomic_numbers()
        ghostz = self.atoms.get_ghost_atomic_numbers()
        return np.concatenate((realz, ghostz))

    def _getradius(self):
        r = PrimiPlotter._getradius(self)
        if len(r) == self.numberofrealatoms + self.numberofghostatoms:
            # Must have calculated radii from atomic numbers
            return r
        else:
            assert len(r) == self.numberofrealatoms
            # Heuristic: use minimum r for the ghosts
            ghostr = min(r) * np.ones(self.numberofghostatoms, float)
            return np.concatenate((r, ghostr))

    def _getcenter(self, coords):
        # max(x) and min(x) only works for rank-1 arrays in Numeric version 17.
        maximal = np.maximum.reduce(coords[:,0:2])
        minimal = np.minimum.reduce(coords[:,0:2])
        self.mpi.max(maximal)
        self.mpi.min(minimal)
        maxx, maxy = maximal
        minx, miny = minimal
        return np.array([maxx + minx, maxy + miny, 0.0]) / 2.0

    def _computevisibility(self, xy, rad, invisible, id, zoom = 1):
        # Find visible atoms, allowing ghost atoms to hide real atoms.
        v = PrimiPlotter._computevisibility(self, xy, rad, invisible, id, zoom)
        # Then remove ghost atoms
        return v * np.less(id, self.numberofrealatoms)

    def _autoscale(self, coords, radii):
        self._verb("Autoscale")
        n = len(self.atoms)
        x = coords[:n,0]
        y = coords[:n,1]
        assert len(x) == len(self.atoms)
        maximal = np.array([max(x), max(y), max(radii[:n])])
        minimal = np.array([min(x), min(y)])
        self.mpi.max(maximal)
        self.mpi.min(minimal)
        maxx, maxy, maxradius = maximal
        minx, miny = minimal
        deltax = maxx - minx + 2*maxradius
        deltay = maxy - miny + 2*maxradius
        scalex = self.dims[0] / deltax
        scaley = self.dims[1] / deltay
        self.scale = 0.95 * min(scalex, scaley)
        self.log("Autoscale: %f" % self.scale)

    def _getcolors(self):
        col = PrimiPlotter._getcolors(self)
        nghost = len(self.atoms.get_ghost_positions())
        newcolshape = (nghost + col.shape[0],) + col.shape[1:]
        newcol = np.zeros(newcolshape, col.dtype)
        newcol[:len(col)] = col
        return newcol
    
    def _makeoutput(self, scale, coords, radii, colors):
        if len(colors.shape) == 1:
            # Greyscales
            ncol = 1
        else:
            ncol = colors.shape[1]  # 1 or 3.
            assert ncol == 3  # RGB values
        # If one processor says RGB, all must convert
        ncolmax = self.mpi.max(ncol)
        if ncolmax > ncol:
            assert ncol == 1
            colors = colors[:,np.newaxis] + np.zeros(ncolmax)[np.newaxis,:]
            ncol = ncolmax
            assert colors.shape == (len(coords), ncol)
        # Now send data from slaves to master
        data = np.zeros((len(coords), 4+ncol), float)
        data[:,:3] = coords
        data[:,3] = radii
        if ncol == 1:
            data[:,4] = colors
        else:
            data[:,4:] = colors
        if not self.master:
            datashape = np.array(data.shape)
            assert datashape.shape == (2,)
            self.mpi.send(datashape, 0, self.mpitag)
            self.mpi.send(data, 0, self.mpitag)
        else:
            total = [data]
            n = len(coords)
            colsmin = colsmax = 4+ncol
            for proc in range(1, self.mpi.size):
                self._verb("Receiving from processor "+str(proc))
                datashape = np.zeros(2, int)
                self.mpi.receive(datashape, proc, self.mpitag)
                fdat = np.zeros(tuple(datashape))
                self.mpi.receive(fdat, proc, self.mpitag)
                total.append(fdat)
                n = n + len(fdat)
                if fdat.shape[1] < colsmin:
                    colsmin = fdat.shape[1]
                if fdat.shape[1] > colsmax:
                    colsmax = fdat.shape[1]
            self._verb("Merging data")
            # Some processors may have only greyscales whereas others
            # may have RGB.  That will cause difficulties.
            trouble = colsmax != colsmin
            data = np.zeros((n, colsmax), float)
            if trouble:
                assert data.shape[1] == 7
            else:
                assert data.shape[1] == 7 or data.shape[1] == 5
            i = 0
            for d in total:
                if not trouble or d.shape[1] == 7:
                    data[i:i+len(d)] = d
                else:
                    assert d.shape[1] == 5
                    data[i:i+len(d), :5] = d
                    data[i:i+len(d), 5] = d[4]
                    data[i:i+len(d), 6] = d[4]
                i = i + len(d)
            assert i == len(data)
            # Now all data is on the master
            self._verb("Sorting merged data")
            order = np.argsort(data[:,2])
            data = data[order]   ### take(data, order)
            coords = data[:,:3]
            radii = data[:,3]
            if data.shape[1] == 5:
                colors = data[:,4]
            else:
                colors = data[:,4:]
            PrimiPlotter._makeoutput(self, scale, coords, radii, colors)
    
class _PostScriptDevice:
    """PostScript based output device."""
    offset = (0,0)   # Will be changed by some classes
    def __init__(self):
        self.scale = 1
        self.linewidth = 1
        self.outline = 1
        
    def set_dimensions(self, dims):
        self.dims = dims

    def set_owner(self, owner):
        self.owner = owner
        
    def inform_about_scale(self, scale):
        self.linewidth = 0.1 * scale

    def set_outline(self, value):
        self.outline = value
        return self   # Can chain these calls in set_output()
        
    def plot(self, *args, **kargs):
        self.Doplot(self.PSplot, *args, **kargs)
        
    def plotArray(self, *args, **kargs):
        self.Doplot(self.PSplotArray, *args, **kargs)
        
    def PSplot(self, file, n, coords, r, colors, noshowpage=0):
        xy = coords[:,:2]
        assert(len(xy) == len(r) and len(xy) == len(colors))
        if len(colors.shape) == 1:
            gray = 1
        else:
            gray = 0
            assert(colors.shape[1] == 3)
        file.write("%!PS-Adobe-2.0\n")
        file.write("%%Creator: Primiplot\n")
        file.write("%%Pages: 1\n")
        file.write("%%%%BoundingBox: %d %d %d %d\n" %
                   (self.offset + (self.offset[0] + self.dims[0],
                                   self.offset[1] + self.dims[1])))
        file.write("%%EndComments\n")
        file.write("\n")
        file.write("% Enforce BoundingBox\n")
        file.write("%d %d moveto %d 0 rlineto 0 %d rlineto -%d 0 rlineto\n" %
                   ((self.offset + self.dims + (self.dims[0],))))
        file.write("closepath clip newpath\n\n")
        file.write("%f %f scale\n" % (2*(1.0/self.scale,)))
        file.write("%d %d translate\n" % (self.scale * self.offset[0],
                                          self.scale * self.offset[1]))
        file.write("\n")
        if gray:
            if self.outline:
                file.write("/circ { 0 360 arc gsave setgray fill grestore stroke } def\n")
            else:
                file.write("/circ { 0 360 arc setgray fill } def\n")
        else:
            if self.outline:
                file.write("/circ { 0 360 arc gsave setrgbcolor fill grestore stroke } def\n")
            else:
                file.write("/circ { 0 360 arc setrgbcolor fill } def\n")
        file.write("%f setlinewidth 0.0 setgray\n" %
                   (self.linewidth * self.scale,))
        
        if gray:
            data = np.zeros((len(xy), 4), float)
            data[:,0] = colors
            data[:,1:3] = (self.scale * xy)
            data[:,3] = (self.scale * r)
            for point in data:
                file.write("%.3f %.2f %.2f %.2f circ\n" % tuple(point))
        else:
            data = np.zeros((len(xy), 6), float)
            data[:,0:3] = colors
            data[:,3:5] = (self.scale * xy)
            data[:,5] = (self.scale * r)
            for point in data:
                file.write("%.3f %.3f %.3f %.2f %.2f %.2f circ\n" % tuple(point))
        if not noshowpage:
            file.write("showpage\n")
            
    def PSplotArray(self, file, n, data, noshowpage=0):
        assert(len(data.shape) == 3)
        assert(data.shape[0] == self.dims[1] and data.shape[1] == self.dims[0])
        data = np.clip((256*data).astype(int), 0, 255)
        file.write("%!PS-Adobe-2.0\n")
        file.write("%%Creator: Fieldplotter\n")
        file.write("%%Pages: 1\n")
        file.write("%%%%BoundingBox: %d %d %d %d\n" %
                   (self.offset + (self.offset[0] + self.dims[0],
                                   self.offset[1] + self.dims[1])))
        file.write("%%EndComments\n")
        file.write("\n")
        file.write("%d %d translate\n" % self.offset)
        file.write("%f %f scale\n" % self.dims)
        file.write("\n")
        file.write("% String holding a single line\n")
        file.write("/pictline %d string def\n" %(data.shape[1]*data.shape[2],))
        file.write("\n")
        file.write("%d %d 8\n" % self.dims)
        file.write("[%d 0 0 %d 0 0]\n" % self.dims)
        file.write("{currentfile pictline readhexstring pop}\n")
        file.write("false %d colorimage\n" % (data.shape[2],))
        file.write("\n")
        s = ""
        for d in data.flat:
            s += ("%02X" % d)
            if len(s) >= 72:
                file.write(s+"\n")
                s = ""
        file.write(s+"\n")
        file.write("\n")
        if not noshowpage:
            file.write("showpage\n")
            
class _PostScriptToFile(_PostScriptDevice):
    """Output device for PS files."""
    compr_suffix = None
    def __init__(self, prefix, compress = 0):
        self.compress = compress
        if "'" in prefix:
            raise ValueError("Filename may not contain a quote ('): "+prefix)
        if "%" in prefix:
            # Assume the user knows what (s)he is doing
            self.filenames = prefix
        else:
            self.filenames = prefix + "%04d" + self.suffix
            if compress:
                if self.compr_suffix is None:
                    raise RuntimeError("Compression not supported.")
                self.filenames = self.filenames + self.compr_suffix
        _PostScriptDevice.__init__(self)

class PostScriptFile(_PostScriptToFile):
    suffix = ".ps"
    compr_suffix = ".gz"
    offset = (50,50)
    # Inherits __init__

    def Doplot(self, plotmethod, n, *args, **kargs):
        filename = self.filenames % (n,)
        self.owner.log("Output to PostScript file "+filename)
        if self.compress:
            file = os.popen("gzip > '"+filename+"'", "w")
        else:
            file = open(filename, "w")
        plotmethod(*(file, n)+args, **kargs)
        file.close()

class _PS_to_bitmap(_PostScriptToFile):
    gscmd = "gs -q -sDEVICE={0} -sOutputFile=- -dDEVICEWIDTH=%d -dDEVICEHEIGHT=%d - "
    # Inherits __init__

    def Doplot(self, plotmethod, n, *args, **kargs):
        filename = self.filenames % (n,)
        self.owner.log("Output to bitmapped file " + filename)
        cmd = self.gscmd.format(self.devicename)
        if self.compress:
            cmd = cmd + "| gzip "
            
        cmd = (cmd+" > '%s'") % (self.dims[0], self.dims[1], filename)
        file = os.popen(cmd, "w")
        plotmethod(*(file, n)+args, **kargs)
        file.close()

class PnmFile(_PS_to_bitmap):
    suffix = ".pnm"
    devicename = "pnmraw"
    compr_suffix = ".gz"

#class GifFile(_PS_via_PnmFile):
#    suffix = ".gif"
#    converter = "| ppmquant -floyd 256 2>/dev/null | ppmtogif 2>/dev/null"

class JpegFile(_PS_to_bitmap):
    suffix = ".jpeg"
    devicename = "jpeg"

class PngFile(_PS_to_bitmap):
    suffix = ".png"
    devicename = "png16m"

class Png256File(_PS_to_bitmap):
    suffix = ".png"
    devicename = "png256"

class X11Window(_PostScriptDevice):
    """Shows the plot in an X11 window."""
    #Inherits __init__
    gscmd = "gs -q -sDEVICE=x11 -dDEVICEWIDTH=%d -dDEVICEHEIGHT=%d -r72x72 -"
    def Doplot(self, plotmethod, n, *args, **kargs):
        self.owner.log("Output to X11 window")
        try:
            file = self.pipe
            self.pipe.write("showpage\n")
        except AttributeError:
            filename = self.gscmd % tuple(self.dims)
            file = os.popen(filename, "w")
            self.pipe = file
        kargs["noshowpage"] = 1
        plotmethod(*(file, n)+args, **kargs)
        file.write("flushpage\n")
        file.flush()

# Helper functions
def _rot(v, axis):
    ax1, ax2 = ((1, 2), (0, 2), (0, 1))[axis]
    c, s = np.cos(v), np.sin(v)
    m = np.zeros((3,3), float)
    m[axis,axis] = 1.0
    m[ax1,ax1] = c
    m[ax2,ax2] = c
    m[ax1,ax2] = s
    m[ax2,ax1] = -s
    return m

def _colorsfromdict(dict, cls):
    """Extract colors from dictionary using cls as key."""
    assert(isinstance(dict, type({})))
    # Allow local modifications, to replace strings with rgb values.
    dict = dict.copy()
    isgray, isrgb = 0, 0
    for k in dict.keys():
        v = dict[k]
        if isinstance(v, basestring):
            v = color_table[v]
            dict[k] = v
        try:
            if len(v) == 3:
                isrgb = 1 # Assume it is an RGB value
                if not hasattr(v, "shape"):
                    dict[k] = np.array(v)   # Convert to array
            else:
                raise RuntimeError("Unrecognized color object "+repr(v))
        except TypeError:
            isgray = 1 # Assume it is a number
    if isgray and isrgb:
        # Convert all to RGB
        for k in dict.keys():
            v = dict[k]
            if not hasattr(v, "shape"):
                dict[k] = v * np.ones(3, float)
    # Now the dictionary is ready
    if isrgb:
        colors = np.zeros((len(cls),3), float)
    else:
        colors = np.zeros((len(cls),), float)
    for i in range(len(cls)):
        colors[i] = dict[cls[i]]
    return colors



"""
This module contains tools for preconditioned geometry optimisation.

Code maintained by James Kermode <james.kermode@gmail.com>
Parts written by John Woolley, Letif Mones and Christoph Ortner.

The preconditioned LBFGS optimizer implemented here is described in
the following publication:

    D. Packwood, J. R. Kermode, L. Mones, N. Bernstein, J. Woolley,
    N. Gould, C. Ortner, and G. Csanyi, A universal preconditioner for
    simulating condensed phase materials, J. Chem. Phys. 144, 164109 (2016).
    DOI: http://dx.doi.org/10.1063/1.4947024

A preconditioned version of FIRE is also included, this is less well tested.

Optional dependencies
---------------------

    - scipy, `pip install scipy` for efficient sparse linear algebra,
      important for large systems (>1000 atoms).
    - PyAMG, `pip install pyamg`, for iterative adaptive multi grid
      invesion of the preconditioner, again important for large systems.
"""


from ase.optimize.precon.precon import Precon, Exp, C1, Pfrommer, FF, Exp_FF
from ase.optimize.precon.lbfgs import PreconLBFGS
from ase.optimize.precon.fire import PreconFIRE

__all__ = ['Precon', 'Exp', 'C1', 'Pfrommer',
           'FF', 'Exp_FF', 'PreconLBFGS', 'PreconFIRE']



import os.path
try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve

url = 'http://wiki.fysik.dtu.dk/ase-files/'


def setup(app):
    pass


for file in ['ase/gui/ag.png',
             'ase/ase-talk.pdf']:
    if os.path.isfile(file):
        continue
    try:
        urlretrieve(url + os.path.basename(file), file)
        print('Downloaded:', file)
    except IOError:
        pass



import sys
import numpy as np
from pymysql import connect
from pymysql.err import ProgrammingError
from copy import deepcopy

from ase.db.sqlite import SQLite3Database
from ase.db.sqlite import init_statements
from ase.db.sqlite import VERSION
from ase.db.postgresql import remove_nan_and_inf, insert_nan_and_inf
import ase.io.jsonio
import json


class Connection(object):
    """
    Wrapper for the MySQL connection

    Arguments
    =========
    host: str
        Hostname. For a local database this is localhost.
    user: str
        Username.
    passwd: str
        Password
    db_name: str
        Name of the database
    port: int
        Port
    binary_prefix: bool
        MySQL checks if an argument can be interpreted as a UTF-8 string. This
        check fails for binary values. Binary values need to have _binary
        prefix in MySQL. By setting this to True, the prefix is automatically
        added for binary values.
    """

    def __init__(self, host=None, user=None, passwd=None, port=3306,
                 db_name=None, binary_prefix=False):
        self.con = connect(host=host, user=user, passwd=passwd, db=db_name,
                           binary_prefix=binary_prefix)

    def cursor(self):
        return MySQLCursor(self.con.cursor())

    def commit(self):
        self.con.commit()

    def close(self):
        self.con.close()


class MySQLCursor(object):
    """
    Wrapper for the MySQL cursor. The most important task performed by this
    class is to translate SQLite queries to MySQL. Translation is needed
    because ASE DB uses some field names that are reserved words in MySQL.
    Thus, these has to mapped onto other field names.
    """
    sql_replace = [
        (' key TEXT', ' attribute_key TEXT'),
        ('(key TEXT', '(attribute_key TEXT'),
        ('SELECT key FROM', 'SELECT attribute_key FROM'),
        ('?', '%s'),
        (' keys ', ' attribute_keys '),
        (' key=', ' attribute_key='),
        ('table.key', 'table.attribute_key'),
        (' IF NOT EXISTS', '')
    ]

    def __init__(self, cur):
        self.cur = cur

    def execute(self, sql, params=None):

        # Replace external table key -> attribute_key
        for substibution in self.sql_replace:
            sql = sql.replace(substibution[0], substibution[1])

        if params is None:
            params = ()

        self.cur.execute(sql, params)

    def fetchone(self):
        return self.cur.fetchone()

    def fetchall(self):
        return self.cur.fetchall()

    def _replace_nan_inf_kvp(self, values):
        for item in values:
            if not np.isfinite(item[1]):
                item[1] = sys.float_info.max / 2
        return values

    def executemany(self, sql, values):
        if 'number_key_values' in sql:
            values = self._replace_nan_inf_kvp(values)

        for substibution in self.sql_replace:
            sql = sql.replace(substibution[0], substibution[1])
        self.cur.executemany(sql, values)


class MySQLDatabase(SQLite3Database):
    """
    ASE interface to a MySQL database (via pymysql package).

    Arguments
    ==========
    url: str
        URL to the database. It should have the form
        mysql://username:password@host:port/database_name.
        Example URL with the following credentials
            username: john
            password: johnspasswd
            host: localhost (i.e. server is running locally)
            database: johns_calculations
            port: 3306
        mysql://john:johnspasswd@localhost:3306/johns_calculations
    create_indices: bool
        Carried over from parent class. Currently indices are not
        created for MySQL, as TEXT fields cannot be hashed by MySQL.
    use_lock_file: bool
        See SQLite
    serial: bool
        See SQLite
    """
    type = 'mysql'
    default = 'DEFAULT'

    def __init__(self, url=None, create_indices=True,
                 use_lock_file=False, serial=False):
        super(MySQLDatabase, self).__init__(
            url, create_indices, use_lock_file, serial)

        self.host = None
        self.username = None
        self.passwd = None
        self.db_name = None
        self.port = 3306
        self._parse_url(url)

    def _parse_url(self, url):
        """
        Parse the URL
        """
        url = url.replace('mysql://', '')
        url = url.replace('mariadb://', '')

        splitted = url.split(':', 1)
        self.username = splitted[0]

        splitted = splitted[1].split('@')
        self.passwd = splitted[0]

        splitted = splitted[1].split('/')
        host_and_port = splitted[0].split(':')
        self.host = host_and_port[0]
        self.port = int(host_and_port[1])
        self.db_name = splitted[1]

    def _connect(self):
        return Connection(host=self.host, user=self.username,
                          passwd=self.passwd, db_name=self.db_name,
                          port=self.port, binary_prefix=True)

    def _initialize(self, con):
        if self.initialized:
            return

        cur = con.cursor()

        information_exists = True
        self._metadata = {}
        try:
            cur.execute("SELECT 1 FROM information")
        except ProgrammingError:
            information_exists = False

        if not information_exists:
            # We need to initialize the DB
            # MySQL require that id is explicitly set as primary key
            # in the systems table
            init_statements_cpy = deepcopy(init_statements)
            init_statements_cpy[0] = init_statements_cpy[0][:-1] + \
                ', PRIMARY KEY(id))'

            statements = schema_update(init_statements_cpy)
            for statement in statements:
                cur.execute(statement)
            con.commit()
            self.version = VERSION
        else:
            cur.execute('select * from information')

            for name, value in cur.fetchall():
                if name == 'version':
                    self.version = int(value)
                elif name == 'metadata':
                    self._metadata = json.loads(value)

        self.initialized = True

    def blob(self, array):
        if array is None:
            return None
        return super(MySQLDatabase, self).blob(array).tobytes()

    def get_offset_string(self, offset, limit=None):
        sql = ''
        if not limit:
            # mysql does not allow for setting limit to -1 so
            # instead we set a large number
            sql += '\nLIMIT 10000000000'
        sql += '\nOFFSET {0}'.format(offset)
        return sql

    def get_last_id(self, cur):
        cur.execute('select max(id) as ID from systems')
        last_id = cur.fetchone()[0]
        return last_id

    def create_select_statement(self, keys, cmps,
                                sort=None, order=None, sort_table=None,
                                what='systems.*'):
        sql, value = super(MySQLDatabase, self).create_select_statement(
            keys, cmps, sort, order, sort_table, what)

        for subst in MySQLCursor.sql_replace:
            sql = sql.replace(subst[0], subst[1])
        return sql, value

    def encode(self, obj, binary=False):
        return ase.io.jsonio.encode(remove_nan_and_inf(obj))

    def decode(self, obj, lazy=False):
        return insert_nan_and_inf(ase.io.jsonio.decode(obj))


def schema_update(statements):
    for i, statement in enumerate(statements):
        for a, b in [('REAL', 'DOUBLE'),
                     ('INTEGER PRIMARY KEY AUTOINCREMENT',
                      'INT NOT NULL AUTO_INCREMENT')]:
            statements[i] = statement.replace(a, b)

    # MySQL does not support UNIQUE constraint on TEXT
    # need to use VARCHAR. The unique_id is generated with
    # randint(16**31, 16**32-1) so it will contain 32
    # hex-characters
    statements[0] = statements[0].replace('TEXT UNIQUE', 'VARCHAR(32) UNIQUE')

    # keys is a reserved word in MySQL redefine this table name to
    # attribute_keys
    statements[2] = statements[2].replace('keys', 'attribute_keys')

    txt2jsonb = ['calculator_parameters', 'key_value_pairs']

    for column in txt2jsonb:
        statements[0] = statements[0].replace(
            '{} TEXT,'.format(column),
            '{} JSON,'.format(column))

    statements[0] = statements[0].replace('data BLOB,', 'data JSON,')

    tab_with_key_field = ['attribute_keys', 'number_key_values',
                          'text_key_values']

    # key is a reserved word in MySQL redefine this to attribute_key
    for i, statement in enumerate(statements):
        for tab in tab_with_key_field:
            if tab in statement:
                statements[i] = statement.replace(
                    'key TEXT', 'attribute_key TEXT')
    return statements



"""
Run VASP tests to ensure that relaxation with the VASP calculator works.
This is conditional on the existence of the VASP_COMMAND or VASP_SCRIPT
environment variables.

"""

from ase.test.vasp import installed2 as installed

assert installed()

import numpy as np
from ase import io
# QuasiNewton nowadays is an alias for BFGSLineSearch, which is
# broken. Use BFGS instead.
from ase.optimize import BFGS as QuasiNewton
from ase.build import bulk
from ase.calculators.vasp import Vasp2 as Vasp

# -- Perform Volume relaxation within Vasp
def vasp_vol_relax():
    Al = bulk('Al', 'fcc', a=4.5, cubic=True)
    calc = Vasp(xc='LDA', isif=7, nsw=5,
                ibrion=1, ediffg=-1e-3, lwave=False, lcharg=False)
    calc.calculate(Al)

    # Explicitly parse atomic position output file from Vasp
    CONTCAR_Al = io.read('CONTCAR', format='vasp')

    print('Stress after relaxation:\n', calc.read_stress())

    print('Al cell post relaxation from calc:\n', calc.get_atoms().get_cell())
    print('Al cell post relaxation from atoms:\n', Al.get_cell())
    print('Al cell post relaxation from CONTCAR:\n', CONTCAR_Al.get_cell())

    # All the cells should be the same.
    assert (calc.get_atoms().get_cell() == CONTCAR_Al.get_cell()).all()
    assert (Al.get_cell() == CONTCAR_Al.get_cell()).all()

    return Al

# -- Perform Volume relaxation using ASE with Vasp as force/stress calculator
def ase_vol_relax():
    Al = bulk('Al', 'fcc', a=4.5, cubic=True)
    calc = Vasp(xc='LDA')
    Al.set_calculator(calc)

    from ase.constraints import StrainFilter
    sf = StrainFilter(Al)
    qn = QuasiNewton(sf, logfile='relaxation.log')
    qn.run(fmax=0.1, steps=5)

    print('Stress:\n', calc.read_stress())
    print('Al post ASE volume relaxation\n', calc.get_atoms().get_cell())

    return Al

# Test function for comparing two cells
def cells_almost_equal(cellA, cellB, tol=0.01):
    return  (np.abs(cellA - cellB) < tol).all()

# Correct LDA relaxed cell
a_rel = 4.18
LDA_cell = np.diag([a_rel, a_rel, a_rel])

Al_vasp = vasp_vol_relax()
Al_ase = ase_vol_relax()

assert cells_almost_equal(LDA_cell, Al_vasp.get_cell())
assert cells_almost_equal(LDA_cell, Al_ase.get_cell())

# Cleanup
Al_ase.get_calculator().clean()



# creates: io1.png io2.png io3.png

from ase import Atoms
from ase.build import fcc111, add_adsorbate
from ase.io import write, read

adsorbate = Atoms('CO')
adsorbate[1].z = 1.1
a = 3.61
slab = fcc111('Cu', (2, 2, 3), a=a, vacuum=7.0)
add_adsorbate(slab, adsorbate, 1.8, 'ontop')

write('io1.png', slab * (3, 3, 1), rotation='10z,-80x')
write('io2.pov', slab * (3, 3, 1), rotation='10z,-80x',
      transparent=False, run_povray=True)
d = a / 2**0.5
write('io3.pov', slab * (2, 2, 1),
      bbox=(d, 0, 3 * d, d * 3**0.5),
      transparent=False, run_povray=True)

write('slab.xyz', slab)
a = read('slab.xyz')
a.get_cell()
a.get_pbc()
write('slab.traj', slab)
b = read('slab.traj')
b.get_cell()
b.get_pbc()



import numpy as np

from ase.calculators.calculator import Calculator
from ase.data import atomic_numbers
from ase.utils import convert_string_to_fd


class SimpleQMMM(Calculator):
    """Simple QMMM calculator."""

    implemented_properties = ['energy', 'forces']

    def __init__(self, selection, qmcalc, mmcalc1, mmcalc2, vacuum=None):
        """SimpleQMMM object.

        The energy is calculated as::

                    _          _          _
            E = E  (R  ) - E  (R  ) + E  (R   )
                 QM  QM     MM  QM     MM  all

        parameters:

        selection: list of int, slice object or list of bool
            Selection out of all the atoms that belong to the QM part.
        qmcalc: Calculator object
            QM-calculator.
        mmcalc1: Calculator object
            MM-calculator used for QM region.
        mmcalc2: Calculator object
            MM-calculator used for everything.
        vacuum: float or None
            Amount of vacuum to add around QM atoms.  Use None if QM
            calculator doesn't need a box.

        """
        self.selection = selection
        self.qmcalc = qmcalc
        self.mmcalc1 = mmcalc1
        self.mmcalc2 = mmcalc2
        self.vacuum = vacuum

        self.qmatoms = None
        self.center = None

        self.name = '{0}-{1}+{1}'.format(qmcalc.name, mmcalc1.name)

        Calculator.__init__(self)

    def initialize_qm(self, atoms):
        constraints = atoms.constraints
        atoms.constraints = []
        self.qmatoms = atoms[self.selection]
        atoms.constraints = constraints
        self.qmatoms.pbc = False
        if self.vacuum:
            self.qmatoms.center(vacuum=self.vacuum)
            self.center = self.qmatoms.positions.mean(axis=0)

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        if self.qmatoms is None:
            self.initialize_qm(atoms)

        self.qmatoms.positions = atoms.positions[self.selection]
        if self.vacuum:
            self.qmatoms.positions += (self.center -
                                       self.qmatoms.positions.mean(axis=0))

        energy = self.qmcalc.get_potential_energy(self.qmatoms)
        qmforces = self.qmcalc.get_forces(self.qmatoms)
        energy += self.mmcalc2.get_potential_energy(atoms)
        forces = self.mmcalc2.get_forces(atoms)

        if self.vacuum:
            qmforces -= qmforces.mean(axis=0)
        forces[self.selection] += qmforces

        energy -= self.mmcalc1.get_potential_energy(self.qmatoms)
        forces[self.selection] -= self.mmcalc1.get_forces(self.qmatoms)

        self.results['energy'] = energy
        self.results['forces'] = forces


class EIQMMM(Calculator):
    """Explicit interaction QMMM calculator."""
    implemented_properties = ['energy', 'forces']

    def __init__(self, selection, qmcalc, mmcalc, interaction,
                 vacuum=None, embedding=None, output=None):
        """EIQMMM object.

        The energy is calculated as::

                    _          _         _    _
            E = E  (R  ) + E  (R  ) + E (R  , R  )
                 QM  QM     MM  MM     I  QM   MM

        parameters:

        selection: list of int, slice object or list of bool
            Selection out of all the atoms that belong to the QM part.
        qmcalc: Calculator object
            QM-calculator.
        mmcalc: Calculator object
            MM-calculator.
        interaction: Interaction object
            Interaction between QM and MM regions.
        vacuum: float or None
            Amount of vacuum to add around QM atoms.  Use None if QM
            calculator doesn't need a box.
        embedding: Embedding object or None
            Specialized embedding object.  Use None in order to use the
            default one.
        output: None, '-', str or file-descriptor.
            File for logging information - default is no logging (None).

        """

        self.selection = selection

        self.qmcalc = qmcalc
        self.mmcalc = mmcalc
        self.interaction = interaction
        self.vacuum = vacuum
        self.embedding = embedding

        self.qmatoms = None
        self.mmatoms = None
        self.mask = None
        self.center = None  # center of QM atoms in QM-box

        self.name = '{0}+{1}+{2}'.format(qmcalc.name,
                                         interaction.name,
                                         mmcalc.name)

        self.output = convert_string_to_fd(output)

        Calculator.__init__(self)

    def initialize(self, atoms):
        self.mask = np.zeros(len(atoms), bool)
        self.mask[self.selection] = True

        constraints = atoms.constraints
        atoms.constraints = []  # avoid slicing of constraints
        self.qmatoms = atoms[self.mask]
        self.mmatoms = atoms[~self.mask]
        atoms.constraints = constraints

        self.qmatoms.pbc = False

        if self.vacuum:
            self.qmatoms.center(vacuum=self.vacuum)
            self.center = self.qmatoms.positions.mean(axis=0)
            print('Size of QM-cell after centering:',
                  self.qmatoms.cell.diagonal(), file=self.output)

        self.qmatoms.calc = self.qmcalc
        self.mmatoms.calc = self.mmcalc

        if self.embedding is None:
            self.embedding = Embedding()

        self.embedding.initialize(self.qmatoms, self.mmatoms)
        print('Embedding:', self.embedding, file=self.output)

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        if self.qmatoms is None:
            self.initialize(atoms)

        self.mmatoms.set_positions(atoms.positions[~self.mask])
        self.qmatoms.set_positions(atoms.positions[self.mask])

        if self.vacuum:
            shift = self.center - self.qmatoms.positions.mean(axis=0)
            self.qmatoms.positions += shift
        else:
            shift = (0, 0, 0)

        self.embedding.update(shift)

        ienergy, iqmforces, immforces = self.interaction.calculate(
            self.qmatoms, self.mmatoms, shift)

        qmenergy = self.qmatoms.get_potential_energy()
        mmenergy = self.mmatoms.get_potential_energy()
        energy = ienergy + qmenergy + mmenergy

        print('Energies: {0:12.3f} {1:+12.3f} {2:+12.3f} = {3:12.3f}'
              .format(ienergy, qmenergy, mmenergy, energy), file=self.output)

        qmforces = self.qmatoms.get_forces()
        mmforces = self.mmatoms.get_forces()

        mmforces += self.embedding.get_mm_forces()

        forces = np.empty((len(atoms), 3))
        forces[self.mask] = qmforces + iqmforces
        forces[~self.mask] = mmforces + immforces

        self.results['energy'] = energy
        self.results['forces'] = forces


def wrap(D, cell, pbc):
    """Wrap distances to nearest neighbor (minimum image convention)."""
    for i, periodic in enumerate(pbc):
        if periodic:
            d = D[:, i]
            L = cell[i]
            d[:] = (d + L / 2) % L - L / 2  # modify D inplace


class Embedding:
    def __init__(self, molecule_size=3, **parameters):
        """Point-charge embedding."""
        self.qmatoms = None
        self.mmatoms = None
        self.molecule_size = molecule_size
        self.virtual_molecule_size = None
        self.parameters = parameters

    def __repr__(self):
        return 'Embedding(molecule_size={0})'.format(self.molecule_size)

    def initialize(self, qmatoms, mmatoms):
        """Hook up embedding object to QM and MM atoms objects."""
        self.qmatoms = qmatoms
        self.mmatoms = mmatoms
        charges = mmatoms.calc.get_virtual_charges(mmatoms)
        self.pcpot = qmatoms.calc.embed(charges, **self.parameters)
        self.virtual_molecule_size = (self.molecule_size *
                                      len(charges) // len(mmatoms))

    def update(self, shift):
        """Update point-charge positions."""
        # Wrap point-charge positions to the MM-cell closest to the
        # center of the the QM box, but avoid ripping molecules apart:
        qmcenter = self.qmatoms.positions.mean(axis=0)
        # if counter ions are used, then molecule_size has more than 1 value
        if self.mmatoms.calc.name == 'combinemm':
            mask1 = self.mmatoms.calc.mask
            mask2 = ~mask1
            vmask1 = self.mmatoms.calc.virtual_mask
            vmask2 = ~vmask1
            apm1 = self.mmatoms.calc.apm1
            apm2 = self.mmatoms.calc.apm2
            spm1 = self.mmatoms.calc.atoms1.calc.sites_per_mol
            spm2 = self.mmatoms.calc.atoms2.calc.sites_per_mol
            pos = self.mmatoms.positions
            pos1 = pos[mask1].reshape((-1, apm1, 3))
            pos2 = pos[mask2].reshape((-1, apm2, 3))
            pos = (pos1, pos2)
        else:
            pos = (self.mmatoms.positions, )
            apm1 = self.molecule_size
            apm2 = self.molecule_size
            # This is only specific to calculators where apm != spm, 
            # i.e. TIP4P. Non-native MM calcs do not have this attr.
            if hasattr(self.mmatoms.calc, 'sites_per_mol'):
                spm1 = self.mmatoms.calc.sites_per_mol
                spm2 = self.mmatoms.calc.sites_per_mol
            else:
                spm1 = self.molecule_size
                spm2 = spm1
            mask1 = np.ones(len(self.mmatoms), dtype=bool)
            mask2 = mask1

        wrap_pos = np.zeros_like(self.mmatoms.positions)
        com_all = []
        apm = (apm1, apm2)
        mask = (mask1, mask2)
        spm = (spm1, spm2)
        for p, n, m, vn in zip(pos, apm, mask, spm):
            positions = p.reshape((-1, n, 3)) + shift

            # Distances from the center of the QM box to the first atom of
            # each molecule:
            distances = positions[:, 0] - qmcenter

            wrap(distances, self.mmatoms.cell.diagonal(), self.mmatoms.pbc)
            offsets = distances - positions[:, 0]
            positions += offsets[:, np.newaxis] + qmcenter

            # Geometric center positions for each mm mol for LR cut
            com = np.array([p.mean(axis=0) for p in positions])
            # Need per atom for C-code:
            com_pv = np.repeat(com, vn, axis=0)
            com_all.append(com_pv)

            wrap_pos[m] = positions.reshape((-1, 3))

        positions = wrap_pos.copy()
        positions = self.mmatoms.calc.add_virtual_sites(positions)

        if self.mmatoms.calc.name == 'combinemm':
            com_pv = np.zeros_like(positions)
            for ii, m in enumerate((vmask1, vmask2)):
                com_pv[m] = com_all[ii]

        # compatibility with gpaw versions w/o LR cut in PointChargePotential
        if 'rc2' in self.parameters:
            self.pcpot.set_positions(positions, com_pv=com_pv)
        else:
            self.pcpot.set_positions(positions)

    def get_mm_forces(self):
        """Calculate the forces on the MM-atoms from the QM-part."""
        f = self.pcpot.get_forces(self.qmatoms.calc)
        return self.mmatoms.calc.redistribute_forces(f)


def combine_lj_lorenz_berthelot(sigmaqm, sigmamm,
                                epsilonqm, epsilonmm):
    """Combine LJ parameters according to the Lorenz-Berthelot rule"""
    sigma = []
    epsilon = []
    # check if input is tuple of vals for more than 1 mm calc, or only for 1.
    if type(sigmamm) == tuple:
        numcalcs = len(sigmamm)
    else:
        numcalcs = 1  # if only 1 mm calc, eps and sig are simply np arrays
        sigmamm = (sigmamm, )
        epsilonmm = (epsilonmm, )
    for cc in range(numcalcs):
        sigma_c = np.zeros((len(sigmaqm), len(sigmamm[cc])))
        epsilon_c = np.zeros_like(sigma_c)

        for ii in range(len(sigmaqm)):
            sigma_c[ii, :] = (sigmaqm[ii] + sigmamm[cc]) / 2
            epsilon_c[ii, :] = (epsilonqm[ii] * epsilonmm[cc])**0.5
        sigma.append(sigma_c)
        epsilon.append(epsilon_c)

    if numcalcs == 1:  # retain original, 1 calc function
        sigma = np.array(sigma[0])
        epsilon = np.array(epsilon[0])

    return sigma, epsilon


class LJInteractionsGeneral:
    name = 'LJ-general'

    def __init__(self, sigmaqm, epsilonqm, sigmamm, epsilonmm,
                 qm_molecule_size, mm_molecule_size=3,
                 rc=np.Inf, width=1.0):
        self.sigmaqm = sigmaqm
        self.epsilonqm = epsilonqm
        self.sigmamm = sigmamm
        self.epsilonmm = epsilonmm
        self.qms = qm_molecule_size
        self.mms = mm_molecule_size
        self.rc = rc
        self.width = width
        self.combine_lj()

    def combine_lj(self):
        self.sigma, self.epsilon = combine_lj_lorenz_berthelot(
            self.sigmaqm, self.sigmamm, self.epsilonqm, self.epsilonmm)

    def calculate(self, qmatoms, mmatoms, shift):
        epsilon = self.epsilon
        sigma = self.sigma

        # loop over possible multiple mm calculators
        # currently 1 or 2, but could be generalized in the future...
        apm1 = self.mms
        mask1 = np.ones(len(mmatoms), dtype=bool)
        mask2 = mask1
        apm = (apm1, )
        sigma = (sigma, )
        epsilon = (epsilon, )
        if hasattr(mmatoms.calc, 'name'):
            if mmatoms.calc.name == 'combinemm':
                mask1 = mmatoms.calc.mask
                mask2 = ~mask1
                apm1 = mmatoms.calc.apm1
                apm2 = mmatoms.calc.apm2
                apm = (apm1, apm2)

        mask = (mask1, mask2)
        e_all = 0
        qmforces_all = np.zeros_like(qmatoms.positions)
        mmforces_all = np.zeros_like(mmatoms.positions)

        # zip stops at shortest tuple so we dont double count
        # cases of no counter ions.
        for n, m, eps, sig in zip(apm, mask, epsilon, sigma):
            mmpositions = self.update(qmatoms, mmatoms[m], n, shift)
            qmforces = np.zeros_like(qmatoms.positions)
            mmforces = np.zeros_like(mmatoms[m].positions)
            energy = 0.0

            qmpositions = qmatoms.positions.reshape((-1, self.qms, 3))

            for q, qmpos in enumerate(qmpositions):  # molwise loop
                # cutoff from first atom of each mol
                R00 = mmpositions[:, 0] - qmpos[0, :]
                d002 = (R00**2).sum(1)
                d00 = d002**0.5
                x1 = d00 > self.rc - self.width
                x2 = d00 < self.rc
                x12 = np.logical_and(x1, x2)
                y = (d00[x12] - self.rc + self.width) / self.width
                t = np.zeros(len(d00))
                t[x2] = 1.0
                t[x12] -= y**2 * (3.0 - 2.0 * y)
                dt = np.zeros(len(d00))
                dt[x12] -= 6.0 / self.width * y * (1.0 - y)
                for qa in range(len(qmpos)):
                    if ~np.any(eps[qa, :]):
                        continue
                    R = mmpositions - qmpos[qa, :]
                    d2 = (R**2).sum(2)
                    c6 = (sig[qa, :]**2 / d2)**3
                    c12 = c6**2
                    e = 4 * eps[qa, :] * (c12 - c6)
                    energy += np.dot(e.sum(1), t)
                    f = t[:, None, None] * (24 * eps[qa, :] *
                         (2 * c12 - c6) / d2)[:, :, None] * R
                    f00 = - (e.sum(1) * dt / d00)[:, None] * R00
                    mmforces += f.reshape((-1, 3))
                    qmforces[q * self.qms + qa, :] -= f.sum(0).sum(0)
                    qmforces[q * self.qms, :] -= f00.sum(0)
                    mmforces[::n, :] += f00

                e_all += energy
                qmforces_all += qmforces
                mmforces_all[m] += mmforces

        return e_all, qmforces_all, mmforces_all

    def update(self, qmatoms, mmatoms, n, shift):
        # Wrap point-charge positions to the MM-cell closest to the
        # center of the the QM box, but avoid ripping molecules apart:
        qmcenter = qmatoms.cell.diagonal() / 2
        positions = mmatoms.positions.reshape((-1, n, 3)) + shift

        # Distances from the center of the QM box to the first atom of
        # each molecule:
        distances = positions[:, 0] - qmcenter

        wrap(distances, mmatoms.cell.diagonal(), mmatoms.pbc)
        offsets = distances - positions[:, 0]
        positions += offsets[:, np.newaxis] + qmcenter

        return positions


class LJInteractions:
    name = 'LJ'

    def __init__(self, parameters):
        """Lennard-Jones type explicit interaction.

        parameters: dict
            Mapping from pair of atoms to tuple containing epsilon and sigma
            for that pair.

        Example:

            lj = LJInteractions({('O', 'O'): (eps, sigma)})

        """
        self.parameters = {}
        for (symbol1, symbol2), (epsilon, sigma) in parameters.items():
            Z1 = atomic_numbers[symbol1]
            Z2 = atomic_numbers[symbol2]
            self.parameters[(Z1, Z2)] = epsilon, sigma
            self.parameters[(Z2, Z1)] = epsilon, sigma

    def calculate(self, qmatoms, mmatoms, shift):
        qmforces = np.zeros_like(qmatoms.positions)
        mmforces = np.zeros_like(mmatoms.positions)
        species = set(mmatoms.numbers)
        energy = 0.0
        for R1, Z1, F1 in zip(qmatoms.positions, qmatoms.numbers, qmforces):
            for Z2 in species:
                if (Z1, Z2) not in self.parameters:
                    continue
                epsilon, sigma = self.parameters[(Z1, Z2)]
                mask = (mmatoms.numbers == Z2)
                D = mmatoms.positions[mask] + shift - R1
                wrap(D, mmatoms.cell.diagonal(), mmatoms.pbc)
                d2 = (D**2).sum(1)
                c6 = (sigma**2 / d2)**3
                c12 = c6**2
                energy += 4 * epsilon * (c12 - c6).sum()
                f = 24 * epsilon * ((2 * c12 - c6) / d2)[:, np.newaxis] * D
                F1 -= f.sum(0)
                mmforces[mask] += f
        return energy, qmforces, mmforces


class RescaledCalculator(Calculator):
    """Rescales length and energy of a calculators to match given
    lattice constant and bulk modulus

    Useful for MM calculator used within a :class:`ForceQMMM` model.
    See T. D. Swinburne and J. R. Kermode, Phys. Rev. B 96, 144102 (2017)
    for a derivation of the scaling constants.
    """
    implemented_properties = ['forces', 'energy', 'stress']

    def __init__(self, mm_calc,
                 qm_lattice_constant, qm_bulk_modulus,
                 mm_lattice_constant, mm_bulk_modulus):
        Calculator.__init__(self)
        self.mm_calc = mm_calc
        self.alpha = qm_lattice_constant / mm_lattice_constant
        self.beta = mm_bulk_modulus / qm_bulk_modulus / (self.alpha**3)

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        # mm_pos = atoms.get_positions()
        scaled_atoms = atoms.copy()

        # scaled_atoms.positions = mm_pos/self.alpha
        mm_cell = atoms.get_cell()
        scaled_atoms.set_cell(mm_cell / self.alpha, scale_atoms=True)

        forces = self.mm_calc.get_forces(scaled_atoms)
        energy = self.mm_calc.get_potential_energy(scaled_atoms)
        stress = self.mm_calc.get_stress(scaled_atoms)

        self.results = {'energy': energy / self.beta,
                        'forces': forces / (self.beta * self.alpha),
                        'stress': stress / (self.beta * self.alpha**3)}


class ForceConstantCalculator(Calculator):
    """
    Compute forces based on provided force-constant matrix

    Useful with `ForceQMMM` to do harmonic QM/MM using force constants
    of QM method.
    """
    implemented_properties = ['forces', 'energy']

    def __init__(self, D, ref, f0):
        """
        Parameters:

        D: matrix or sparse matrix, shape `(3*len(ref), 3*len(ref))`
            Force constant matrix.
            Sign convention is `D_ij = d^2E/(dx_i dx_j), so
            `force = -D.dot(displacement)`
        ref: ase.atoms.Atoms
            Atoms object for reference configuration
        f0: array, shape `(len(ref), 3)`
            Value of forces at reference configuration
        """
        assert D.shape[0] == D.shape[1]
        assert D.shape[0] // 3 == len(ref)
        self.D = D
        self.ref = ref
        self.f0 = f0
        self.size = len(ref)
        Calculator.__init__(self)

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)
        u = atoms.positions - self.ref.positions
        f = -self.D.dot(u.reshape(3 * self.size))
        forces = np.zeros((len(atoms), 3))
        forces[:, :] = f.reshape(self.size, 3)
        self.results['forces'] = forces + self.f0
        self.results['energy'] = 0.0


class ForceQMMM(Calculator):
    """
    Force-based QM/MM calculator

    QM forces are computed using a buffer region and then mixed abruptly
    with MM forces:

        F^i_QMMM = {   F^i_QM    if i in QM region
                   {   F^i_MM    otherwise

    cf. N. Bernstein, J. R. Kermode, and G. Csanyi,
    Rep. Prog. Phys. 72, 026501 (2009)
    and T. D. Swinburne and J. R. Kermode, Phys. Rev. B 96, 144102 (2017).
    """
    implemented_properties = ['forces', 'energy']

    def __init__(self,
                 atoms,
                 qm_selection_mask,
                 qm_calc,
                 mm_calc,
                 buffer_width,
                 vacuum=5.,
                 zero_mean=True):
        """
        ForceQMMM calculator

        Parameters:

        qm_selection_mask: list of ints, slice object or bool list/array
            Selection out of atoms that belong to the QM region.
        qm_calc: Calculator object
            QM-calculator.
        mm_calc: Calculator object
            MM-calculator (should be scaled, see :class:`RescaledCalculator`)
            Can use `ForceConstantCalculator` based on QM force constants, if
            available.
        vacuum: float or None
            Amount of vacuum to add around QM atoms.
        zero_mean: bool
            If True, add a correction to zero the mean force in each direction
        """

        if len(atoms[qm_selection_mask]) == 0:
            raise ValueError("no QM atoms selected!")

        self.qm_selection_mask = qm_selection_mask
        self.qm_calc = qm_calc
        self.mm_calc = mm_calc
        self.vacuum = vacuum
        self.buffer_width = buffer_width
        self.zero_mean = zero_mean

        self.qm_buffer_mask = None
        self.cell = None
        self.qm_shift = None

        Calculator.__init__(self)

    def initialize_qm_buffer_mask(self, atoms):
        """
        Initialises system to perform qm calculation
        """

        # get the radius of the qm_selection in non periodic directions
        qm_positions = atoms[self.qm_selection_mask].get_positions()
        # identify qm radius as an larges distance from the center
        # of the cluster (overestimation)
        qm_center = qm_positions.mean(axis=0)

        non_pbc_directions = np.logical_not(self.atoms.pbc)

        centered_positions = atoms.get_positions()

        for i, non_pbc in enumerate(non_pbc_directions):
            if non_pbc:
                qm_positions.T[i] -= qm_center[i]
                centered_positions.T[i] -= qm_center[i]

        qm_radius = np.linalg.norm(qm_positions.T, axis=1).max()
        self.cell = self.atoms.cell.copy()

        for i, non_pbc in enumerate(non_pbc_directions):
            if non_pbc:
                self.cell[i][i] = 2.0 * (qm_radius +
                                         self.buffer_width +
                                         self.vacuum)

        # identify atoms in region < qm_radius + buffer
        distances_from_center = np.linalg.norm(
            centered_positions.T[non_pbc_directions].T, axis=1)

        self.qm_buffer_mask = (distances_from_center <
                               qm_radius + self.buffer_width)

        # exclude atoms that are too far (in case of non spherical region)
        for i, buffer_atom in enumerate(self.qm_buffer_mask &
                                        np.logical_not(self.qm_selection_mask)):
            if buffer_atom:
                distance = np.linalg.norm(
                    (qm_positions -
                     centered_positions[i]).T[non_pbc_directions].T, axis=1)
                if distance.min() > self.buffer_width:
                    self.qm_buffer_mask[i] = False

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        if self.qm_buffer_mask is None:
            self.initialize_qm_buffer_mask(atoms)

        # initialize the object
        # qm_buffer_atoms = atoms.copy()
        qm_buffer_atoms = atoms[self.qm_buffer_mask]
        del qm_buffer_atoms.constraints

        qm_buffer_atoms.set_cell(self.cell)
        qm_shift = (0.5 * qm_buffer_atoms.cell.diagonal() -
                    qm_buffer_atoms.positions.mean(axis=0))

        qm_buffer_atoms.set_cell(self.cell)
        qm_buffer_atoms.positions += qm_shift

        forces = self.mm_calc.get_forces(atoms)

        qm_forces = self.qm_calc.get_forces(qm_buffer_atoms)
        forces[self.qm_selection_mask] = \
            qm_forces[self.qm_selection_mask[self.qm_buffer_mask]]

        if self.zero_mean:
            # Target is that: forces.sum(axis=1) == [0., 0., 0.]
            forces[:] -= forces.mean(axis=0)

        self.results['forces'] = forces
        self.results['energy'] = 0.0




"""
Run some tests to ensure that VASP calculator constructs correct POTCAR files

"""

from os import remove
from os.path import isfile
from ase.atoms import Atoms
from ase.calculators.vasp import Vasp
from ase.test.vasp import installed

assert installed()


def check_potcar(setups, filename='POTCAR'):
    """Return true if labels in setups are found in POTCAR"""

    pp = []
    with open(filename, 'r') as f:
        for line in f:
            if 'TITEL' in line.split():
                pp.append(line.split()[3])
    for setup in setups:
        assert setup in pp

# Write some POTCARs and check they are ok
potcar = 'POTCAR'
try:
    atoms = Atoms('CaGdCs',
                  positions=[[0, 0, 1], [0, 0, 2], [0, 0, 3]], cell=[5, 5, 5])

    calc = Vasp(xc='pbe')
    calc.initialize(atoms)
    calc.write_potcar()
    check_potcar(('Ca_pv', 'Gd', 'Cs_sv'), filename=potcar)

    calc = Vasp(xc='pbe', setups='recommended')
    calc.initialize(atoms)
    calc.write_potcar()
    check_potcar(('Ca_sv', 'Gd_3', 'Cs_sv'), filename=potcar)

    calc = Vasp(xc='pbe', setups='materialsproject')
    calc.initialize(atoms)
    calc.write_potcar()
    check_potcar(('Ca_sv', 'Gd', 'Cs_sv'), filename=potcar)

    atoms = Atoms('CaInI',
                  positions=[[0, 0, 1], [0, 0, 2], [0, 0, 3]], cell=[5, 5, 5])
    calc = Vasp(xc='pbe', setups={'base': 'gw'})
    calc.initialize(atoms)
    calc.write_potcar()
    check_potcar(('Ca_sv_GW', 'In_d_GW', 'I_GW'), filename=potcar)

    calc = Vasp(xc='pbe', setups={'base': 'gw', 'I': ''})
    calc.initialize(atoms)
    calc.write_potcar()
    check_potcar(('Ca_sv_GW', 'In_d_GW', 'I'), filename=potcar)

    calc = Vasp(xc='pbe', setups={'base': 'gw', 'Ca': '_sv', 2: 'I'})
    calc.initialize(atoms)
    calc.write_potcar()
    check_potcar(('Ca_sv', 'In_d_GW', 'I'), filename=potcar)
finally:
    if isfile(potcar):
        remove(potcar)



import json


class CLICommand:
    """Get calculations from NOMAD and write to JSON files.

    ...
    """

    @staticmethod
    def add_arguments(p):
        p.add_argument('uri', nargs='+', metavar='nmd://<hash>',
                       help='URIs to get')

    @staticmethod
    def run(args):
        from ase.nomad import download
        for uri in args.uri:
            calculation = download(uri)
            identifier = calculation.hash.replace('/', '.')
            fname = 'nmd.{}.nomad-json'.format(identifier)
            with open(fname, 'w') as fd:
                json.dump(calculation, fd)
            print(uri)



# flake8: noqa

import numpy as np

import ase
import ase.lattice.hexagonal
from ase.build import bulk, molecule

from ase.neighborlist import (mic, neighbor_list, primitive_neighbor_list,
                              first_neighbors)

tol = 1e-7

# two atoms
a = ase.Atoms('CC', positions=[[0.5, 0.5, 0.5], [1,1,1]], cell=[10, 10, 10],
              pbc=True)
i, j, d = neighbor_list("ijd", a, 1.1)
assert (i == np.array([0, 1])).all()
assert (j == np.array([1, 0])).all()
assert np.abs(d - np.array([np.sqrt(3/4), np.sqrt(3/4)])).max() < tol

# test_neighbor_list
for pbc in [True, False, [True, False, True]]:
    a = ase.Atoms('4001C', cell=[29, 29, 29])
    a.set_scaled_positions(np.transpose([np.random.random(len(a)),
                                         np.random.random(len(a)),
                                         np.random.random(len(a))]))
    j, dr, i, abs_dr, shift = neighbor_list("jDidS", a, 1.85)

    assert (np.bincount(i) == np.bincount(j)).all()

    r = a.get_positions()
    dr_direct = mic(r[j]-r[i], a.cell)
    assert np.abs(r[j]-r[i]+shift.dot(a.cell) - dr_direct).max() < tol

    abs_dr_from_dr = np.sqrt(np.sum(dr*dr, axis=1))
    abs_dr_direct = np.sqrt(np.sum(dr_direct*dr_direct, axis=1))

    assert np.all(np.abs(abs_dr-abs_dr_from_dr) < 1e-12)
    assert np.all(np.abs(abs_dr-abs_dr_direct) < 1e-12)

    assert np.all(np.abs(dr-dr_direct) < 1e-12)

# test_neighbor_list_atoms_outside_box
for pbc in [True, False, [True, False, True]]:
    a = ase.Atoms('4001C', cell=[29, 29, 29])
    a.set_scaled_positions(np.transpose([np.random.random(len(a)),
                                         np.random.random(len(a)),
                                         np.random.random(len(a))]))
    a.set_pbc(pbc)
    a.positions[100, :] += a.cell[0, :]
    a.positions[200, :] += a.cell[1, :]
    a.positions[300, :] += a.cell[2, :]
    j, dr, i, abs_dr, shift = neighbor_list("jDidS", a, 1.85)

    assert (np.bincount(i) == np.bincount(j)).all()

    r = a.get_positions()
    dr_direct = mic(r[j]-r[i], a.cell)
    assert np.abs(r[j]-r[i]+shift.dot(a.cell) - dr_direct).max() < tol

    abs_dr_from_dr = np.sqrt(np.sum(dr*dr, axis=1))
    abs_dr_direct = np.sqrt(np.sum(dr_direct*dr_direct, axis=1))

    assert np.all(np.abs(abs_dr-abs_dr_from_dr) < 1e-12)
    assert np.all(np.abs(abs_dr-abs_dr_direct) < 1e-12)

    assert np.all(np.abs(dr-dr_direct) < 1e-12)

# test_small_cell
a = ase.Atoms('C', positions=[[0.5, 0.5, 0.5]], cell=[1, 1, 1],
              pbc=True)
i, j, dr, shift = neighbor_list("ijDS", a, 1.1)
assert np.bincount(i)[0] == 6
assert (dr == shift).all()

i, j = neighbor_list("ij", a, 1.5)
assert np.bincount(i)[0] == 18

a.set_pbc(False)
i = neighbor_list("i", a, 1.1)
assert len(i) == 0

a.set_pbc([True, False, False])
i = neighbor_list("i", a, 1.1)
assert np.bincount(i)[0] == 2

a.set_pbc([True, False, True])
i = neighbor_list("i", a, 1.1)
assert np.bincount(i)[0] == 4

# test_out_of_cell_small_cell
a = ase.Atoms('CC', positions=[[0.5, 0.5, 0.5],
                               [1.1, 0.5, 0.5]],
              cell=[1, 1, 1], pbc=False)
i1, j1, r1 = neighbor_list("ijd", a, 1.1)
a.set_cell([2, 1, 1])
i2, j2, r2 = neighbor_list("ijd", a, 1.1)

assert (i1 == i2).all()
assert (j1 == j2).all()
assert np.abs(r1 - r2).max() < tol

# test_out_of_cell_large_cell
a = ase.Atoms('CC', positions=[[9.5, 0.5, 0.5],
                               [10.1, 0.5, 0.5]],
              cell=[10, 10, 10], pbc=False)
i1, j1, r1 = neighbor_list("ijd", a, 1.1)
a.set_cell([20, 10, 10])
i2, j2, r2 = neighbor_list("ijd", a, 1.1)

assert (i1 == i2).all()
assert (j1 == j2).all()
assert np.abs(r1 - r2).max() < tol

# test_hexagonal_cell
for sx in range(3):
    a = ase.lattice.hexagonal.Graphite('C', latticeconstant=(2.5, 10.0),
                                       size=[sx+1,sx+1,1])
    i = neighbor_list("i", a, 1.85)
    assert np.all(np.bincount(i)==3)

# test_first_neighbors
i = [1,1,1,1,3,3,3]
assert (first_neighbors(5, i) == np.array([0,0,4,4,7,7])).all()
i = [0,1,2,3,4,5]
assert (first_neighbors(6, i) == np.array([0,1,2,3,4,5,6])).all()

# test_multiple_elements
a = molecule('HCOOH')
a.center(vacuum=5.0)
i = neighbor_list("i", a, 1.85)
assert (np.bincount(i) == np.array([2,3,1,1,1])).all()

cutoffs = {(1, 6): 1.2}
i = neighbor_list("i", a, cutoffs)
assert (np.bincount(i) == np.array([0,1,0,0,1])).all()

cutoffs = {(6, 8): 1.4}
i = neighbor_list("i", a, cutoffs)
assert (np.bincount(i) == np.array([1,2,1])).all()

cutoffs = {('H', 'C'): 1.2, (6, 8): 1.4}
i = neighbor_list("i", a, cutoffs)
assert (np.bincount(i) == np.array([1,3,1,0,1])).all()

cutoffs = [0.0, 0.9, 0.0, 0.5, 0.5]
i = neighbor_list("i", a, cutoffs)
assert (np.bincount(i) == np.array([0,1,0,0,1])).all()

cutoffs = [0.7, 0.9, 0.7, 0.5, 0.5]
i = neighbor_list("i", a, cutoffs)
assert (np.bincount(i) == np.array([2,3,1,1,1])).all()

# test_noncubic
a = bulk("Al", cubic=False)
i, j, d = neighbor_list("ijd", a, 3.1)
assert (np.bincount(i) == np.array([12])).all()
assert np.abs(d - [2.86378246]*12).max() < tol

# test pbc
nat = 10
atoms = ase.Atoms(numbers=range(nat),
                  cell=[(0.2, 1.2, 1.4),
                        (1.4, 0.1, 1.6),
                        (1.3, 2.0, -0.1)])
atoms.set_scaled_positions(3 * np.random.random((nat, 3)) - 1)

for p1 in range(2):
    for p2 in range(2):
        for p3 in range(2):
            atoms.set_pbc((p1, p2, p3))
            i, j, d, D, S = neighbor_list("ijdDS", atoms, atoms.numbers * 0.2 + 0.5)
            c = np.bincount(i, minlength=len(atoms))
            atoms2 = atoms.repeat((p1 + 1, p2 + 1, p3 + 1))
            i2, j2, d2, D2, S2 = neighbor_list("ijdDS", atoms2, atoms2.numbers * 0.2 + 0.5)
            c2 = np.bincount(i2, minlength=len(atoms))
            c2.shape = (-1, nat)
            dd = d.sum() * (p1 + 1) * (p2 + 1) * (p3 + 1) - d2.sum()
            dr = np.linalg.solve(atoms.cell.T, (atoms.positions[1]-atoms.positions[0]).T).T+np.array([0,0,3])
            assert abs(dd) < 1e-10
            assert not (c2 - c).any()

c = 0.0058
i, j, d = primitive_neighbor_list('ijd',
                                  [True, True, True],
                                  np.eye(3) * 7.56,
                                  np.array([[0, 0, 0],
                                            [0, 0, 0.99875]]),
                                  [c, c],
                                  self_interaction=False,
                                  use_scaled_positions=True)
assert np.all(i == [0, 1])
assert np.all(j == [1, 0])
assert np.allclose(d, [0.00945, 0.00945])

# Empty atoms object
i, D, d, j, S = neighbor_list("iDdjS", ase.Atoms(), 1.0)
assert i.dtype == np.int
assert j.dtype == np.int
assert d.dtype == np.float
assert D.dtype == np.float
assert S.dtype == np.int
assert i.shape == (0,)
assert j.shape == (0,)
assert d.shape == (0,)
assert D.shape == (0, 3)
assert S.shape == (0, 3)

# Check that only a scalar (not a tuple) is returned if we request a single
# argument.
i = neighbor_list("i", ase.Atoms(), 1.0)
assert i.dtype == np.int
assert i.shape == (0,)



"""Function-like object creating monoclinic lattices.

The following lattice creator is defined:
    SimpleMonoclinic
    BaseCenteredMonoclinic
"""

from ase.lattice.triclinic import TriclinicFactory
import numpy as np


class SimpleMonoclinicFactory(TriclinicFactory):
    "A factory for creating simple monoclinic lattices."
    # The name of the crystal structure in ChemicalElements
    xtal_name = "monoclinic"

    def make_crystal_basis(self):
        "Make the basis matrix for the crystal unit cell and the system unit cell."
        # First convert the basis specification to a triclinic one
        if isinstance(self.latticeconstant, type({})):
            self.latticeconstant['beta'] = 90
            self.latticeconstant['gamma'] = 90
        else:
            if len(self.latticeconstant) == 4:
                self.latticeconstant = self.latticeconstant + (90,90)
            else:
                raise ValueError("Improper lattice constants for monoclinic crystal.")

        TriclinicFactory.make_crystal_basis(self)
        
SimpleMonoclinic = SimpleMonoclinicFactory()

class BaseCenteredMonoclinicFactory(SimpleMonoclinicFactory):
    # The natural basis vectors of the crystal structure
    int_basis = np.array([[1, -1, 0],
                          [1, 1, 0],
                          [0, 0, 2]])
    basis_factor = 0.5

    # Converts the natural basis back to the crystallographic basis
    inverse_basis = np.array([[1, 1, 0],
                              [-1, 1, 0],
                              [0, 0, 1]])
    inverse_basis_factor = 1.0

BaseCenteredMonoclinic = BaseCenteredMonoclinicFactory()
